"""Use this as a template of how to make your board compatible with the 
NeuriGUI.
Requirement is that you maintain everything inside the template and only 
add your own steps to the existing ones. This is because the GUI has strict
methods of importing classes and fucntions as well as sharing variables 
between frontend and backend.
"""


from json import dumps
from time import time
from numpy import expand_dims, concatenate, append, array, zeros, reshape
try:
    from io_manager import IOManager
except:
    from .io_manager import IOManager
    

class SamplingUtilsMuseSInteraxon():

    def __init__(self):
        pass # Necessary init function because we call attributes of this class in process_samples()
    

    def message_to_samples(self, lsl_message_eeg, s_chans):
        # -----------------------------------------------------------------
        # Extract samples from board
        # Input
        #   str_message (str) Raw message string coming from board
        #   s_chans     (int) How many channels shall the samples be 
        #               assigned to
        # Default in case of faulty message
        eeg_array       = expand_dims(
            array([0] * s_chans, dtype=float), # Crucial to set float
            axis=1)
        eeg_valid       = False
        #   eeg_array   Numpy array (floats), being channels by samples
        #       (Has to be 2D even if only one channel present [samples x 1])
        #   eeg_valid   Boolean whether workable data or not
        # -----------------------------------------------------------------
        
        eeg_array[:,0] = lsl_message_eeg[0]
        eeg_valid = True
        
        return eeg_array, eeg_valid


    def process_samples(self, transmitter, parameter, shared_buffer,
                     shared_timestamp, gui_running):
        """
        parameter = {
            "baud_rate":    parameter.baud_rate,
            "time_out":     parameter.time_out,
            "port":         parameter.port,
            "start_code":   parameter.start_code, # Start code that boards reads as sign to startt sampling (int)
            "max_chans":    parameter.max_chans, # Specifies how many channels the incoming signals has (int)
            "buffer_length":parameter.buffer_length, # Length of constructed signal buffer (int)
            "buffer_add":   parameter.buffer_add, # Length of padding to avoid filter artifacts  (int)
            "sample_rate":  parameter.sample_rate, # Sampling rate (Hz, int)
            "PGA":          parameter.PGA, # Programmable gain (int)
            "saving_interval": parameter.saving_interval, # how often the data gets written to disk (seconds, int)
            "udp_ip":       parameter.udp_ip, # IP the data shall be forwarded to
            "udp_port":     parameter.udp_port, # Port of signal forwarding
            "set_customsession": parameter.set_customsession, # Whether or not to have a custom output file name (bool)
            "sessionName":  parameter.sessionName, # Custom file name for output (str)
        }"""

        io_manager = IOManager(parameter)
        receiver_eeg, receiver_ppg = io_manager.set_up_lsl_entry_point()

        if parameter["streamed_data_type"] == "EEG":
            receiver = receiver_eeg
        elif parameter["streamed_data_type"] == "PPG":
            receiver = receiver_ppg

        # Prealloate values of loop ---------------------------------------
        start_time          = int(time() * 1000)
        time_stamp_now      = int(time() * 1000) # Do NOT copy from start_time (will generate pointer)
        time_reset          = int(time() * 1000) # Do NOT copy from start_time (will generate pointer)
        sample_count        = int(0)
        
        buffer              = zeros((parameter["max_chans"],
            (parameter["buffer_length"] + parameter["buffer_add"]) * parameter["sample_rate"]))
        time_stamps         = zeros(parameter["buffer_length"] * parameter["sample_rate"], dtype=int)
        pga                 = parameter["PGA"]
        s_chans             = parameter["max_chans"]
        sampling_rate       = parameter["sample_rate"]
        saving_interval     = parameter["saving_interval"] * parameter["sample_rate"]
        update_buffer       = zeros((buffer.shape[0], buffer.shape[1]+1))
        update_times        = zeros((buffer.shape[1]+1), dtype=int)

        # Preallocate json relay message and relay connection
        relay_array         = io_manager.build_standard_relay_message(parameter["max_chans"])
        udp_ip              = parameter["udp_ip"]
        udp_port            = parameter["udp_port"]
            
        while gui_running.value == 1:
            
            # =========================================================
            # ========= B O A R D   S P E C I F I C   P A R T =========
            # =========================================================
            raw_message_eeg        = receiver.pull_sample()

            # TODO: Circumvent fact that PPG is sampled at 64 Hz and
            # slows down EEG sampling which is sending at 256 Hz.
            # raw_message_ppg        = receiver_ppg.pull_sample()

            # TODO: Adapt message_to_samples() in order to convert
            # raw_message string into numpy array of samples
            buffer_in, valid_eeg    = self.message_to_samples(raw_message_eeg, s_chans)

            if not valid_eeg:
                # TODO: Implement an interpolation system
                continue                
            # =========================================================
            # =========================================================
            # =========================================================

            # Standard set of procedures from here on: You should not
            # require any additional changes from here

            # Current timestamp -------------------------------------------
            time_stamp_now          = int(raw_message_eeg[1] / 1000) - start_time
            # This will generate unchanged time_stamps for all samples of 
            # the incoming buffer (= 10 in case of bluetooth), but that is
            # not a problem

            sample_count        = sample_count + 1
            
            sample              = buffer_in[:, 0]

            for iSample in range(s_chans):
                relay_array["".join(["c", str(iSample+1)])] = str(sample[iSample])

            update_buffer       = concatenate((buffer, expand_dims(sample, 1)), axis=1)
            update_times        = append(time_stamps, time_stamp_now)

            # Build new buffer and timestamp arrays
            buffer              = update_buffer[:, 1:]
            time_stamps         = update_times[1:]

            # Make data available for downstream programs
            transmitter.sendto(bytes(dumps(relay_array), "utf-8"), (udp_ip, udp_port))

            # Update shared memory allocations for frontend
            shared_buffer[:] = reshape(buffer, buffer.size) # Has left edge for filtering
            shared_timestamp.value = time_stamp_now

            # Write out samples to file -----------------------------------
            if sample_count == saving_interval:

                io_manager.calc_sample_rate(time_stamp_now, time_reset, 
                                    sampling_rate, time_stamps)

                io_manager.master_write_data(buffer, time_stamps, 
                    saving_interval)
                sample_count        = 0
                time_reset          = time_stamp_now

        return


from scipy.signal import butter, lfilter_zi, lfilter, hilbert
from numpy import zeros, pad, abs


class Processing():

    def __init__(self, parameters):

        self.pm                 = parameters
        self.prepare_filters()


    def prepare_filters(self):

        # Bandpass filters
        # -----------------------------------------------------------------
        self.b_detrend, self.a_detrend          = butter(
            self.pm.filter_order, self.pm.frequency_bands["Whole"][0],
            btype='highpass', fs=self.pm.sample_rate)
        self.b_slow, self.a_slow                = butter(
            self.pm.filter_order, self.pm.frequency_bands["Slow"],
            btype='bandpass', fs=self.pm.sample_rate)
        
        if self.pm.streamed_data_type == "EEG":
            self.b_wholerange, self.a_wholerange    = butter(
                self.pm.filter_order, self.pm.frequency_bands["Whole"],
                btype='bandpass', fs=self.pm.sample_rate)
            self.b_sleep, self.a_sleep              = butter(
                self.pm.filter_order, self.pm.frequency_bands["Sleep"],
                btype='bandpass', fs=self.pm.sample_rate)
            self.b_theta, self.a_theta              = butter(
                self.pm.filter_order, self.pm.frequency_bands["Theta"],
                btype='bandpass', fs=self.pm.sample_rate)
            self.b_notch, self.a_notch              = butter(
                self.pm.filter_order, self.pm.frequency_bands["LineNoise"],
                btype='bandstop', fs=self.pm.sample_rate)
            self.b_notch60, self.a_notch60          = butter(
                self.pm.filter_order, self.pm.frequency_bands["LineNoise60"],
                btype='bandstop', fs=self.pm.sample_rate)

        # Determine padding length for signal filtering
        # -----------------------------------------------------------------
        default_pad     = 6 * max(len(self.a_detrend), len(self.b_detrend))
        if default_pad > self.pm.buffer_length * self.pm.sample_rate/10-1:
            self.padlen = int(default_pad) # Scipy expects int
        else:
            self.padlen = int(self.pm.buffer_length*self.pm.sample_rate/10-1) # Scipy expects int


    def filter_signal(self, signal, b, a):
        # =================================================================
        # Input:
        #   signal              Numpy 1D array [samples]
        # Output:
        #   signal_filtered[0]  1D numpy array of filtered signal where 
        #                       first sample is 0
        # =================================================================
        padded_signal   = pad(signal, (self.padlen, 0), 'symmetric')
        init_state      = lfilter_zi(b, a) # 1st sample --> 0
        signal_filtered = lfilter(b, a, padded_signal, 
            zi=init_state*padded_signal[0])
        signal_filtered = signal_filtered[0][self.padlen:]
        return signal_filtered


    def extract_envelope(self, signal):
        v_hilbert       = signal
        for iChan in range(signal.shape[0]):
            # padded_signal   = np.pad(signal[iChan,], (self.padlen, self.padlen), 'symmetric')
            # hilbert[iChan,] = np.abs(scipy.signal.hilbert(padded_signal))[self.padlen:-self.padlen]
            v_hilbert[iChan,] = abs(hilbert(signal[iChan,]))
        return v_hilbert


    def downsample(self, buffer, s_down):
        # =================================================================
        # Input:
        #   buffer              Numpy array [channels x samples]
        # Output:
        #   downsamples_buffer  Numpy array of downsampled signal, same  
        #                       dimensions as input buffer
        # =================================================================

        downsampled_signal  = zeros((buffer.shape[0], int(buffer.shape[1]/s_down)))
        idx_retain = range(0, buffer.shape[1], s_down)
        for iChan in range(self.pm.max_chans):
            # downsampled_signal[iChan,] = scipy.signal.decimate(buffer[iChan,], s_down)
            downsampled_signal[iChan,] = buffer[iChan,idx_retain]

        return downsampled_signal


    def prepare_buffer(self, buffer, bSB, aSB, bPB, aPB):
        # =================================================================
        # Input:
        #   buffer              Numpy array [channels x samples]
        #   bSB, aSB            Filter coefficients as put out by 
        #                       scipy.signal.butter (Stopband)
        #   bPB, aPB            Filter coefficients as put out by 
        #                       scipy.signal.butter (Passband)
        # Output:
        #   filtered_buffer     Numpy array of filtered signal, same  
        #                       dimensions as input buffer
        # =================================================================
        x_shape             = buffer.shape

        noise_free_signal   = zeros(x_shape)
        filtered_buffer     = zeros(x_shape)
        for iChan in range(x_shape[0]):

            # Reject ambiant electrical noise (at 50 Hz)
            # -------------------------------------------------------------
            if all(bSB != None):
                noise_free_signal[iChan,] = self.filter_signal(
                    buffer[iChan,], bSB, aSB)
            else:
                noise_free_signal[iChan,] = buffer[iChan,]

            # Extract useful frequency range
            # -------------------------------------------------------------
            if all(bPB != None):
                filtered_buffer[iChan,] = self.filter_signal(
                    noise_free_signal[iChan,], bPB, aPB)
            else:
                filtered_buffer[iChan,] = noise_free_signal[iChan,]

        return filtered_buffer


import tkinter                              as tk
from PIL                                    import Image, ImageTk
from sys                                    import platform
from github                                 import Github
import serial.tools.list_ports
import customtkinter
import os
import socket
import webbrowser
try:
    from backend.compatible_boards          import COMPATIBLE_BOARDS
except:
    from ..backend.compatible_boards        import COMPATIBLE_BOARDS

class Parameters:

    def __init__(self):

        super(Parameters, self).__init__()

        self.frontend_path  = os.path.dirname(__file__)

        self.conf_file      = os.path.join(".", "settings.cfg")
        self.version        = '2.91.4' # TO-DO: Find a more elegant way to dynamically define the current version as this line here gets forgotten a lot
        self.ico_neuri      = os.path.join(self.frontend_path, "Neuri_logo.ico")

        self.set_defaults() # Necessary to execute first in case user 
                            # parameter not found in configuration file
        # Loop up user settings
        self.load_parameters()

        # Build relay connection for other programs
        self.build_relay(self.udp_ip)

        self.get_screen_info()
        self.build_frontend()

        # This stops code execution until paramWin closed
        self.paramWin.mainloop()


    def build_relay(self, ip):
        # =================================================================
        # This connection will be used in order to transfer the incoming 
        # signal from the board to a dynamuically defined port via UDP
        # =================================================================

        self.udp_port   = self.search_free_com(ip)
        self.udp_ip     = ip
        self.send_sock  = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP
        print('Relay connection established at ' + self.udp_ip + ':' + str(self.udp_port))
        print('Use this connection to import signals in your own program!\n')
    

    def search_free_com(self, ip):

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        for iPort in range(12344, 12350):
            try:
                s = s.connect((ip, iPort))
            except:
                return iPort

        raise Exception('No available UDP port found for signal relay')


    def load_parameters(self):

        if not os.path.exists(self.conf_file):
            return
        
        with open(self.conf_file, 'r') as f:
            
            settings                        = f.readlines()
            
            for i, setting in enumerate(settings):
                if 'Darkmode' in setting and 'True' in setting:
                    self.darkmode           = True
                elif 'Darkmode' in setting and 'False' in setting:
                    self.darkmode           = False
                if 'Headless' in setting and 'True' in setting:
                    self.run_headless       = True
                elif 'Headless' in setting and 'False' in setting:
                    self.run_headless       = False
                elif 'SamplingRate' in setting:
                    try:
                        self.sample_rate    = int(setting[setting.find('=')+1:])
                    except:
                        print("Could not load \"SamplingRate\" from configuration")
                elif 'AmountChannels' in setting:
                    try:
                        self.max_chans      = int(setting[setting.find('=')+1:])
                    except:
                        print("Could not load \"AmountChannels\" from configuration")
                elif 'TimeRange' in setting:
                    try:
                        self.buffer_length  = int(setting[setting.find('=')+1:])
                    except:
                        print("Could not load \"TimeRange\" from configuration")
                elif 'PGA' in setting:
                    try:
                        self.PGA            = int(setting[setting.find('=')+1:])
                    except:
                        print("Could not load \"PGA\" from configuration")
                elif 'Port' in setting:
                    try:
                        self.port           = str(setting[setting.find('=')+1:])
                        if '\n' in self.port:
                            self.port = self.port.replace('\n', '')
                        ports = [port.device for port in list(serial.tools.list_ports.comports())]
                        if len([p for p in ports if p == self.port]) == 0:
                            self.port = ''
                    except:
                        print("Could not load \"Port\" from configuration")
                elif 'Board' in setting:
                    try:
                        self.board          = str(setting[setting.find('=')+1:])
                        if '\n' in self.board:
                            self.board = self.board.replace('\n', '')
                    except:
                        print("Could not get board information from configuration")
                elif 'DownsamplingFactor' in setting:
                    try:
                        self.s_down         = int(setting[setting.find('=')+1:])
                    except:
                        print("Could not load \"DownsamplingFactor\" from configuration")

        print("Loaded user-defined settings")

        # Missing: self.selected_chans


    def save_parameters(self):

        end_line = "\n"

        if not os.path.exists(self.conf_file): # File generation

            with open(self.conf_file, 'w') as f:

                settings = [
                    "".join(["Darkmode=", str(self.darkmode), end_line]),
                    "".join(["SamplingRate=", str(self.sample_rate), end_line]),
                    "".join(["AmountChannels=", str(self.max_chans), end_line]),
                    "".join(["TimeRange=", str(self.buffer_length), end_line]),
                    "".join(["PGA=", str(self.PGA), end_line]),
                    "".join(["Port=", str(self.port), end_line]),
                    "".join(["Board=", str(self.board), end_line]),
                    "".join(["DownsamplingFactor=", str(self.s_down), end_line]),
                    "".join(["Headless=", str(self.run_headless), end_line])
                    ]

                f.write("".join(settings))

        else:

            with open(self.conf_file, 'r') as f:
            
                settings                        = f.readlines()
                
                for i, setting in enumerate(settings): # Update values
                    if 'Darkmode' in setting:
                        settings[i]             = "".join(["Darkmode=", str(self.darkmode), end_line])
                    elif 'SamplingRate' in setting:
                        settings[i]             = "".join(["SamplingRate=", str(self.sample_rate), end_line])
                    elif 'AmountChannels' in setting:
                        settings[i]             = "".join(["AmountChannels=", str(self.max_chans), end_line])
                    elif 'TimeRange' in setting:
                        settings[i]             = "".join(["TimeRange=", str(self.buffer_length), end_line])
                    elif 'PGA' in setting:
                        settings[i]             = "".join(["PGA=", str(self.PGA), end_line])
                    elif 'Port' in setting:
                        settings[i]             = "".join(["Port=", str(self.port), end_line])
                    elif 'Board' in setting:
                        settings[i]             = "".join(["Board=", str(self.board), end_line])
                    elif 'DownsamplingFactor' in setting:
                        settings[i]             = "".join(["DownsamplingFactor=", str(self.s_down), end_line])
                    elif 'Headless' in setting:
                        settings[i]             = "".join(["Headless=", str(self.run_headless), end_line])

                new_settings = []
                if len([s for s in settings if "Darkmode" in s]) == 0:
                    new_settings.append("".join(["Darkmode=", str(self.darkmode), end_line]))
                if len([s for s in settings if "SamplingRate" in s]) == 0:
                    new_settings.append("".join(["SamplingRate=", str(self.sample_rate), end_line]))
                if len([s for s in settings if "AmountChannels" in s]) == 0:
                    new_settings.append("".join(["AmountChannels=", str(self.max_chans), end_line]))
                if len([s for s in settings if "TimeRange" in s]) == 0:
                    new_settings.append("".join(["TimeRange=", str(self.buffer_length), end_line]))
                if len([s for s in settings if "PGA" in s]) == 0:
                    new_settings.append("".join(["PGA=", str(self.PGA), end_line]))
                if len([s for s in settings if "Port" in s]) == 0:
                    new_settings.append("".join(["Port=", str(self.port), end_line]))
                if len([s for s in settings if "Board" in s]) == 0:
                    new_settings.append("".join(["Board=", str(self.board), end_line]))
                if len([s for s in settings if "DownsamplingFactor" in s]) == 0:
                    new_settings.append("".join(["DownsamplingFactor=", str(self.s_down), end_line]))
                if len([s for s in settings if "Headless" in s]) == 0:
                    new_settings.append("".join(["Headless=", str(self.run_headless), end_line]))

            with open(self.conf_file, 'w') as f:
                f.write("".join(settings + new_settings))


    def set_defaults(self):

        self.all_set        = False
        self.darkmode       = False
        self.run_headless   = False

        #Session-specific parameters
        self.yrange         = [-0, 0] # List of scalars ([negative, positive]) in order to set figure y axis range
        self.notch          = 50 # Integer 0 (Off), 50 (50 Hz) or 60 (60 Hz)
        self.bpass          = 0 # Integer -1 to 3 according to number of options in "frequency_bands" below
        self.dispenv        = False # Boolean 0 (Off), 1 (On)
        self.set_customsession = False

        #Signal arrays
        self.sample_rate    = COMPATIBLE_BOARDS["Neuri V1 by Helment"][1] #Hertz
        self.max_chans      = COMPATIBLE_BOARDS["Neuri V1 by Helment"][4] #scalar (Max. amount of input channels of board)
        self.selected_chans = [True] * self.max_chans
        self.buffer_length  = 10 #scalar (seconds)
        self.buffer_add     = 4 #scalar (seconds), we add this to the buffer for filtering to avoid edge artifacts
        self.saving_interval= 1 #scalar (seconds)
        self.PGA            = COMPATIBLE_BOARDS["Neuri V1 by Helment"][9] #scalar

        #Signal reception
        self.baud_rate      = COMPATIBLE_BOARDS["Neuri V1 by Helment"][2] #scalar default baudrate for connection
        self.port           = '' #Leave blank
        self.board          = '' #Leave blank
        self.start_code     = COMPATIBLE_BOARDS["Neuri V1 by Helment"][0]
        self.uses_lsl       = COMPATIBLE_BOARDS["Neuri V1 by Helment"][5]
        self.adjustable_pga = COMPATIBLE_BOARDS["Neuri V1 by Helment"][6]
        self.secondary_sampling_rate = COMPATIBLE_BOARDS["Neuri V1 by Helment"][7]
        self.secondary_max_chans = COMPATIBLE_BOARDS["Neuri V1 by Helment"][8]
        self.streamed_data_type = "EEG"
        self.time_out       = None #Wait for message

        # Signal relay
        self.udp_ip         = "127.0.0.1" # Loopback ip for on-device communication

        #Plotting
        # self.plot_intv       = 200 #scalar defining update rate of figure (ms) OBSOLETE PARAMETER
        self.s_down         = 1 #Desired downsampling factor (buffer_length*sample_rate/s_down must be convertable to integer)


        #Signal processing
        self.filter_order   = 3 #scalar
        self.frequency_bands= {
            'LineNoise':    (48, 52),
            'LineNoise60':  (58, 62),
            'Sleep':        (1, 30),
            'Theta':        (4, 8),
            'Whole':        (0.5, 45),
            'Slow':         (0.5, 6),
        }


    def build_frontend(self):

        self.framePadX          = 20
        self.framePadY          = 10
        self.widgetPadX         = 10
        self.widgetPadY         = 20
        
        # Build GUI
        # -----------------------------------------------------------------
        self.paramWin           = customtkinter.CTk()
        pixels_x, pixels_y      = int(
            round(0.8*self.screen_width)), int(round(0.9*self.screen_height))
        x_cordinate, y_cordinate= int((self.screen_width/2) - (pixels_x/2)), int(0)
        self.paramWin.geometry("{}x{}+{}+{}".format(
            pixels_x, pixels_y, x_cordinate, y_cordinate))

        if self.darkmode:
            customtkinter.set_appearance_mode("dark")
        else:
            customtkinter.set_appearance_mode("light")
        customtkinter.set_default_color_theme("blue")

        # Decorations
        self.paramWin.title('Neuri GUI: Settings')
        photo = ImageTk.PhotoImage(Image.open(self.ico_neuri))
        self.paramWin.wm_iconphoto(True, photo)
        

        # Add options
        self.display_version(self.paramWin)

        # Elements to place into srollable frame
        frameScroll = self.add_scrollable_frame(self.paramWin)
        self.display_board_version(self.add_frame_ext_x(frameScroll))
        self.display_ports(self.add_frame_ext_x(frameScroll))
        
        self.display_gains(self.add_frame_ext_x(frameScroll))
        self.display_data_type(self.add_frame_ext_x(frameScroll))
        self.display_samplingrate(self.add_frame_ext_x(frameScroll))
        self.display_timerange(self.add_frame_ext_x(frameScroll))
        self.display_channels(self.add_frame_ext_x(frameScroll))
        self.display_output_name(self.add_frame_ext_x(frameScroll))
        # self.display_speed_up(self.add_frame_ext_x(frameScroll))
        self.display_headless(self.add_frame_ext_x(frameScroll))
        self.display_validate(self.paramWin)
        
        # Set GUI interaction behavior
        self.paramWin.lift()
        self.paramWin.attributes("-topmost", True)
        self.paramWin.after_idle(self.paramWin.attributes, '-topmost', False)

        # Set closing sequence
        # -----------------------------------------------------------------
        # Just closing the window makes the sampling process hang. We 
        # prevent this by setting up a protocol.
        self.paramWin.protocol('WM_DELETE_WINDOW', self.on_closing)


    def add_frame_ext_x(self, master):
        frameMain               = customtkinter.CTkFrame(master=master)
        frameMain.pack(pady=self.framePadY, padx=self.framePadX,
                       fill="both", expand=True)
        return frameMain


    def add_scrollable_frame(self, master):
        frameScroll             = customtkinter.CTkScrollableFrame(
            master=master, bg_color="transparent", fg_color="transparent")
        frameScroll.pack(pady=self.framePadY, padx=self.framePadX,
                         fill="both", expand=True)
        return frameScroll
    

    def display_version(self, master):

        frameVersion             = customtkinter.CTkFrame(
            master=master, bg_color="transparent", fg_color="transparent")
        frameVersion.pack(pady=0, padx=self.framePadX, fill=tk.X,
                          expand=False, side=tk.TOP)

        repository = "SoftcodingForYou/NeuriGUI"

        g = Github()
        
        try:
            latest_release = g.get_repo(repository).get_latest_release()
            v = latest_release.title.replace("V","")
            v = v.replace("v","")
            v = v.replace(".", "")
            ver_latest = int(v)

            thisv = int(self.version.replace(".", ""))

            if thisv < ver_latest:

                target_webpage = "https://github.com/SoftcodingForYou/NeuriGUI/releases/tag/{}".format(latest_release.title)

                button = customtkinter.CTkButton(
                    frameVersion, text="Download latest version",
                    command=lambda: self.open_webpage(target_webpage))
                button.pack(
                        pady=self.widgetPadY, padx=self.widgetPadX,
                        fill=tk.BOTH, expand=False, side=tk.RIGHT)
                
                customtkinter.CTkLabel(master=frameVersion, 
                justify=customtkinter.RIGHT,
                text="".join(["Realease version ",
                              str(latest_release.title.replace("V","")),
                              " available"]),
                text_color='red').pack(
                    pady=self.widgetPadY, padx=self.widgetPadX,
                    fill=tk.BOTH, expand=False, side=tk.RIGHT)
        except:
            pass

        customtkinter.CTkLabel(master=frameVersion, 
            justify=customtkinter.RIGHT,
            text="".join(["GUI version: ", self.version])).pack(
                pady=self.widgetPadY, padx=self.widgetPadX,
                fill=tk.BOTH, expand=False, side=tk.RIGHT)
        

    def open_webpage(self, target):
        webbrowser.open(target)
        
    
    def get_screen_info(self):

        root = tk.Tk()
        # Get information about screen to center the windows
        self.screen_width = root.winfo_screenwidth()
        self.screen_height = root.winfo_screenheight()
        root.destroy()


    def get_board_features(self, board_name):
        self.start_code = COMPATIBLE_BOARDS[board_name][0]
        self.sample_rate = COMPATIBLE_BOARDS[board_name][1]
        self.baud_rate = COMPATIBLE_BOARDS[board_name][2]
        self.max_chans = COMPATIBLE_BOARDS[board_name][4]
        self.selected_chans = [True] * self.max_chans
        self.uses_lsl = COMPATIBLE_BOARDS[board_name][5]
        self.adjustable_pga = COMPATIBLE_BOARDS[board_name][6]
        self.secondary_sampling_rate = COMPATIBLE_BOARDS[board_name][7]
        self.secondary_max_chans = COMPATIBLE_BOARDS[board_name][8]
        self.PGA = COMPATIBLE_BOARDS[board_name][9]

    
    def display_board_version(self, master):
        
        if self.board == '': 
            defaultBoard = list(COMPATIBLE_BOARDS.keys())[0]
            self.board   = defaultBoard
        else:
            defaultBoard = self.board

        self.get_board_features(defaultBoard)

        labelBoard = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Select board')
        labelBoard.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)
        BoardMenu = customtkinter.CTkOptionMenu(master,
                                                values=list(COMPATIBLE_BOARDS.keys()),
                                               command=self.select_board)
        BoardMenu.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)
        BoardMenu.set(defaultBoard)


    def set_menu_states(self):

        try:
            if self.uses_lsl:
                self.port   = "Lab Streaming Layer"
                self.portMenu.set(self.port)
                self.portMenu.configure(state="disabled")
            else:
                ports = [port.device for port in list(serial.tools.list_ports.comports())]
                if self.port != "":
                    self.portMenu.set(self.port) if len(ports) > 0 else self.portMenu.set('No port available')
                    self.portMenu.configure(state="enabled")
                else:
                    self.portMenu.set(ports[0]) if len(ports) > 0 else self.portMenu.set('No port available')
                    self.portMenu.configure(state="enabled")
        except AttributeError:
            pass

        try:
            if self.adjustable_pga:
                self.gainMenu.set(self.PGA)
                self.gainMenu.configure(state="enabled")
            else:
                self.gainMenu.configure(state="disabled")
        except AttributeError:
            pass

        try:
            self.data_type_val.set(value="EEG")
        except AttributeError:
            pass


    def select_board(self, event):

        self.board       = str(event)
        print('Board set to {}'.format(self.board))

        for _, board_name in enumerate(list(COMPATIBLE_BOARDS.keys())):
            if board_name == self.board:
                self.get_board_features(board_name)
                self.predefine_sampling_rate(self.sample_rate)
                self.select_channel_amount(self.max_chans)
                self.set_menu_states()
                return
        raise Exception("Could not pre-define parameters for selected board")


    def display_ports(self, master):

        ports = [port.device for port in list(serial.tools.list_ports.comports())]
        if len(ports) == 0:
            defaultPort = 'No port available'
        else:
            if self.port == '' and not self.uses_lsl:
                defaultPort = ports[0]
                self.port   = defaultPort # If no change, select_port does not get triggered
            elif self.uses_lsl:
                defaultPort = "Lab Streaming Layer"
            else:
                defaultPort = self.port

        labelPort = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Select port')
        labelPort.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)
        self.portMenu = customtkinter.CTkOptionMenu(master, values=ports,
                                               command=self.select_port)
        self.portMenu.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)
        self.portMenu.set(defaultPort)

        self.set_menu_states()

        infoText = """
            You seem to use a GNU/Linux-based OS. Make sure you are part of
            the "dialout" and "uucp" groups by running:
                > sudo usermod -a -G dialout $USER
                > sudo usermod -a -G uucp $USER
            in the terminal in order to use the system's ports without root.
        """

        if platform == "linux" or platform == "linux2":
            labelLinux = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text=infoText)
            labelLinux.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.RIGHT)


    def select_port(self, event):

        self.port       = event
        print('Port set to {}'.format(self.port))


    # def display_protocol(self, master):

    #     labelProt = customtkinter.CTkLabel(master=master, 
    #                                         justify=customtkinter.LEFT,
    #                                         text='Select connection protocol')
    #     labelProt.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)

    #     self.prot_var  = tk.DoubleVar(value=self.neuri_start_signal)

    #     protocolKeys = list(self.neuri_start_signals.keys())
    #     protocolVals = list(self.neuri_start_signals.values())

    #     for i in range(len(self.neuri_start_signals)):
    #         rb = customtkinter.CTkRadioButton(master=master,
    #                                           variable=self.prot_var,
    #                                           value=protocolVals[i],
    #                                           text=str(protocolKeys[i]),
    #                                           command=self.select_protocol)
    #         rb.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)


    # def select_protocol(self):
        
    #     self.firmfeedback = self.prot_var.get()
    #     print('Will send a \"{}\" to board'.format(self.firmfeedback))


    def display_data_type(self, master):

        labelDataType = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Select data type to stream')
        labelDataType.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)
        
        data_type_vals = ["EEG", "PPG"]
        self.data_type_val = tk.StringVar(value=data_type_vals[0], name=data_type_vals[0])

        self.data_type_radio_buttons = []
        for i in range(len(data_type_vals)):
            self.data_type_radio_buttons.append(customtkinter.CTkRadioButton(master=master,
                                              variable=self.data_type_val,
                                              value=data_type_vals[i],
                                              text=str(data_type_vals[i]),
                                              command=self.select_data_type))
            self.data_type_radio_buttons[i].pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)


    def select_data_type(self):
        
        if self.data_type_val.get() == "EEG" or (not self.secondary_max_chans or not self.secondary_sampling_rate):
            if not self.secondary_max_chans or not self.secondary_sampling_rate:
                print("Board does not allow for secondary data type. Board will stream main data type.")
            self.streamed_data_type = "EEG"
            self.max_chans = COMPATIBLE_BOARDS[self.board][4]
            self.sample_rate = COMPATIBLE_BOARDS[self.board][1]
            self.data_type_val.set(value="EEG")
        elif self.data_type_val.get() == "PPG":
            self.streamed_data_type = "PPG"
            self.max_chans = COMPATIBLE_BOARDS[self.board][8]
            self.sample_rate = COMPATIBLE_BOARDS[self.board][7]
        
        self.predefine_sampling_rate(self.sample_rate)
        self.select_channel_amount(self.max_chans)
        print('Board will send {} data'.format(self.streamed_data_type))


    def display_gains(self, master):

        gains = ['1', '2', '4', '6', '8', '12', '24']
        if not self.PGA:
            idx_def = [-1]
        else:
            idx_def = [i for i in range(len(gains)) if int(gains[i]) == self.PGA]

        labelGain = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Select gain')
        labelGain.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)
        self.gainMenu = customtkinter.CTkOptionMenu(master, values=gains,
                                                    command=self.select_gain)
        self.gainMenu.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)
        self.gainMenu.set(gains[int(idx_def[0])])

        self.set_menu_states()


    def select_gain(self, event):

        self.PGA = int(event)
        print('PGA set to {}'.format(self.PGA))


    def display_samplingrate(self, master):
        
        self.labelSfr = customtkinter.CTkLabel(master=master, 
            justify=customtkinter.LEFT,
            text='Optional: Set sampling rate')
        self.labelSfr.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)

        self.masterSfr              = customtkinter.CTkFrame(master=master, bg_color="transparent", fg_color="transparent")
        self.masterSfr.pack(pady=0, padx=0, fill="both", expand=True, side=tk.RIGHT)
        self.frameSfr              = customtkinter.CTkFrame(master=self.masterSfr, bg_color="transparent", fg_color="transparent")
        self.frameSfr.pack(pady=0, padx=0, fill="both", expand=True, side=tk.RIGHT)

        self.labelInfo = customtkinter.CTkLabel(master=self.frameSfr, 
            justify=customtkinter.LEFT,
            text='Caution: Setting wrong values will corrupt data visualization')
        self.labelInfo.pack(pady=(0, self.widgetPadY), padx=self.widgetPadX, side=tk.BOTTOM)

        self.textSfr = customtkinter.CTkEntry(master=self.frameSfr,
                                           width=200, height=15,
                                           placeholder_text=str(self.sample_rate))
        self.textSfr.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, fill=tk.X, expand=True)

        self.buttonValidate = customtkinter.CTkButton(master=self.frameSfr,
                                                 command=self.select_sampling_rate,
                                                 text='Validate')
        self.buttonValidate.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)


    def select_sampling_rate(self):

        try:
            newSR = int(self.textSfr.get())

            if newSR >= 100 or self.streamed_data_type != "EEG":
                self.sample_rate = newSR
                self.labelInfo.configure(text='Sampling rate set to {} Hz'.format(self.sample_rate),
                                        text_color='green')
                print('Sampling rate set to {} Hz.'.format(self.sample_rate))
            else:
                print('Sampling rate must be at least 100 Hz. Reverting back to {} Hz'.format(self.sample_rate))
                self.labelInfo.configure(text='Sampling rate must be at least 100 Hz. Reverting back to {} Hz'.format(self.sample_rate),
                                        text_color='red')
                self.textSfr.configure(placeholder_text='Sampling rate changed to ' + str(self.sample_rate))
        except:
            print('Please enter an integer value. Reverting back to {} Hz'.format(self.sample_rate))
            self.labelInfo.configure(text='Please enter an integer value. Reverting back to {} Hz'.format(self.sample_rate),
                                     text_color='red')


    def predefine_sampling_rate(self, sfr):
        self.sample_rate = sfr

        self.frameSfr.destroy()
        self.frameSfr              = customtkinter.CTkFrame(master=self.masterSfr, bg_color="transparent", fg_color="transparent")
        self.frameSfr.pack(pady=0, padx=0, fill="both", expand=True, side=tk.RIGHT)

        self.labelInfo = customtkinter.CTkLabel(master=self.frameSfr, 
            justify=customtkinter.LEFT,
            text='Changed to default sampling rate of selected board: {} Hz.'.format(self.sample_rate))
        self.labelInfo.pack(pady=(0, self.widgetPadY), padx=self.widgetPadX, side=tk.BOTTOM)

        self.textSfr = customtkinter.CTkEntry(master=self.frameSfr,
                                           width=200, height=15,
                                           placeholder_text=str(self.sample_rate))
        self.textSfr.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, fill=tk.X, expand=True)

        self.buttonValidate = customtkinter.CTkButton(master=self.frameSfr,
                                                 command=self.select_sampling_rate,
                                                 text='Validate')
        self.buttonValidate.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)

        print('Changed to default sampling rate of selected board: {} Hz.'.format(self.sample_rate))


    def display_timerange(self, master):

        labelRange = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Select time range to display')
        labelRange.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)
        
        ranges          = [5, 10, 20]
        idx_def         = [i for i in range(len(ranges)) if int(ranges[i]) == self.buffer_length]

        self.range_var = customtkinter.IntVar(value=ranges[int(idx_def[0])])

        for i in range(len(ranges)):
            rb = customtkinter.CTkRadioButton(master=master,
                                              variable=self.range_var,
                                              value=ranges[i],
                                              text=str(ranges[i]) + ' s',
                                              command=self.select_timerange)
            rb.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)


    def select_timerange(self):
        
        self.buffer_length = self.range_var.get()
        print('Time range to display set to {} s'.format(self.buffer_length))


    def display_channels(self, master):

        labelChans = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Channels to display\n(Only affects the visualization)')
        labelChans.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)

        self.channelInfo = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.RIGHT,
                                            text="")
        self.channelInfo.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.RIGHT)

        self.masterChannels              = customtkinter.CTkFrame(master=master, bg_color="transparent", fg_color="transparent")
        self.masterChannels.pack(pady=0, padx=0, fill="both", expand=True, side=tk.RIGHT)
        self.frameChannels              = customtkinter.CTkFrame(master=self.masterChannels, bg_color="transparent", fg_color="transparent")
        self.frameChannels.pack(pady=0, padx=0, fill="both", expand=True, side=tk.RIGHT)
        self.channels = []
        for i in range(self.max_chans):
            self.channels.append(tk.BooleanVar())
            cb = customtkinter.CTkCheckBox(master=self.frameChannels,
                                            text='Ch. ' + str(i+1),
                                            command=self.select_channels,
                                            variable=self.channels[i],
                                            onvalue=True,
                                            offvalue=False,
                                            width=50)
            cb.select()
            cb.pack(pady=self.widgetPadY, padx=0, side=tk.LEFT, expand=True, fill=None)


    def select_channel_amount(self, numchans):

        self.max_chans      = numchans
        self.selected_chans = [True] * self.max_chans
        self.frameChannels.destroy()
        self.frameChannels              = customtkinter.CTkFrame(master=self.masterChannels, bg_color="transparent", fg_color="transparent")
        self.frameChannels.pack(pady=0, padx=0, fill="both", expand=True, side=tk.RIGHT)
        self.channels       = []
        for i in range(self.max_chans):
            self.channels.append(tk.BooleanVar())
            cb = customtkinter.CTkCheckBox(master=self.frameChannels,
                                            text='Ch. ' + str(i+1),
                                            command=self.select_channels,
                                            variable=self.channels[i],
                                            onvalue=True,
                                            offvalue=False)
            cb.select()
            cb.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)
        self.channelInfo.configure(text="")


    def select_channels(self):

        for i, channel in enumerate(self.channels):
            if self.selected_chans[i] != channel.get():
                self.selected_chans[i] = channel.get()
                if channel.get():
                    print('Channel {} enabled'.format(i))
                else:
                    print('Channel {} disabled'.format(i))

        if not any(self.selected_chans):
            self.channelInfo.configure(text="Select at least one channel",
                text_color='red')
        else:
            self.channelInfo.configure(text="")


    def display_output_name(self, master):
        
        self.sessionName = 'Neuri_[timestamp]'
        self.labelSession = customtkinter.CTkLabel(master=master, 
            justify=customtkinter.CENTER,
            text='Optional: Set a session name')
        self.labelSession.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)

        self.sessionInfo = customtkinter.CTkLabel(master=master, 
            justify=customtkinter.LEFT,
            text='')
        self.sessionInfo.pack(pady=(0, self.widgetPadY), padx=self.widgetPadX, side=tk.BOTTOM)

        self.textSession = customtkinter.CTkEntry(master=master,
                                           width=250, height=15,
                                           placeholder_text='Default: ' + self.sessionName)
        self.textSession.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, fill=tk.X, expand=True)

        self.sessionValidate = customtkinter.CTkButton(master=master,
                                                 command=self.select_output_name,
                                                 text='Validate')
        self.sessionValidate.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)


    def select_output_name(self):

        newSessionName      = str(self.textSession.get())

        if len(newSessionName) > 0:
            self.sessionName    = newSessionName
            self.sessionName    = self.sessionName.replace('\n', '')
            self.set_customsession = True
            self.sessionInfo.configure(text='Session name changed to ' + str(self.sessionName),
                             text_color='green')
            print('Session named \"{}\".'.format(self.sessionName))
        else:
            self.sessionInfo.configure(text='Custom name can not be empty',
                             text_color='red')
            print('Session name can not be empty')


    def display_speed_up(self, master):

        down_facotr = ['1', '2', '5', '10'] # TO-DO: Fix GUI freezing when chosing '1' (no downsampling)
        idx_def     = [i for i in range(len(down_facotr)) if int(down_facotr[i]) == self.s_down]

        labelSpeed  = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Optional: Select downsampling intensity')
        labelSpeed.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)
        SpeedMenu   = customtkinter.CTkOptionMenu(master, values=down_facotr,
                                                    command=self.select_speed_up)
        SpeedMenu.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)
        SpeedMenu.set(down_facotr[int(idx_def[0])])

        infoText    = """
            Selecting high amounts of channels and long time ranges to display can have impacts on the performance.
            You might try to compensate this by setting a higher downsampling factor. This will downsample the data
            for visualization (not the recorded data itself)
            """

        labelInfo   = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text=infoText)
        labelInfo.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.RIGHT)


    def select_speed_up(self, event):

        self.s_down = int(event)
        print('Downsampling intensity set to {}'.format(self.s_down))


    def display_headless(self, master):

        labelHeadless = customtkinter.CTkLabel(master=master, 
                                            justify=customtkinter.LEFT,
                                            text='Sample data without GUI (headless)')
        labelHeadless.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT)

        self.toggled_headless = tk.BooleanVar()
        cb = customtkinter.CTkCheckBox(master=master,
                                        text='',
                                        command=self.set_headless,
                                        variable=self.toggled_headless,
                                        onvalue=True,
                                        offvalue=False)
        if self.run_headless:
            cb.select()
        cb.pack(pady=self.widgetPadY, padx=self.widgetPadX, side=tk.LEFT, expand=True)


    def set_headless(self):

        self.run_headless = self.toggled_headless.get()


    def display_validate(self, master):

        customtkinter.CTkFrame(master, height=1, bg_color='transparent').pack(side=tk.LEFT, fill=tk.X)
        customtkinter.CTkFrame(master, height=1, bg_color='transparent').pack(side=tk.RIGHT, fill=tk.X)

        customtkinter.CTkButton(master,
            text="Save & Start",
            border_width=2,
            border_color="#3B8ED0",
            text_color="#3B8ED0",
            fg_color="white",
            hover_color= "#B8C8D4",
            command=self.on_validating).pack(
                pady=self.framePadY, padx=self.framePadX, side=tk.LEFT)
        
        customtkinter.CTkButton(master,
            text="Reset & Quit",
            border_width=2,
            border_color="#c54040",
            text_color="#c54040",
            fg_color="white",
            hover_color= "#EDC5C5",
            command=self.on_reset).pack(
                pady=self.framePadY, padx=self.framePadX, side=tk.RIGHT)


    def on_validating(self):
        
        print("Parameters set")
        self.save_parameters()
        print("".join(["Parameters stored in ", self.conf_file]))
        self.paramWin.destroy()
        print("Starting GUI now...")


    def on_reset(self):
        
        self.set_defaults()
        print("Parameters have been reset")
        self.save_parameters()
        print("".join(["Settings in ", self.conf_file, "have been reset"]))
        print("Please start the GUI again")
        quit()


    def on_closing(self):

        print("Chose to quit program")
        quit()


from PyQt5                                  import QtCore, QtWidgets, QtGui
from PIL                                    import Image, ImageTk
from tkinter                                import ttk
from sys                                    import platform
import tkinter                              as tk

class Aux:

    def __init__(self):
        self.img_neuri    = './frontend/Neuri_logo.png'
        self.get_screen_info()


    def get_screen_info(self):
        root = tk.Tk()
        # Get information about screen to center the windows
        self.screen_width = root.winfo_screenwidth()
        self.screen_height = root.winfo_screenheight()
        root.destroy()


    def disp_pyqt_splash(self):
        splash_pix = QtGui.QPixmap(self.img_neuri)
        splash_pix = splash_pix.scaled(500, 500, QtCore.Qt.KeepAspectRatio)
        splash = QtWidgets.QSplashScreen(splash_pix, QtCore.Qt.WindowStaysOnTopHint)
        splash.show()

    
    def disp_splash(self):
        root = tk.Tk()

        # multiple image size by zoom
        pixels_x, pixels_y = tuple([int(0.10 * x) for x in Image.open(self.img_neuri).size])
        img = ImageTk.PhotoImage(Image.open(self.img_neuri).resize((pixels_x, pixels_y)), master=root)

        x_cordinate = int((self.screen_width/2) - (pixels_x/2))
        y_cordinate = int((self.screen_height/2) - (pixels_y/2))

        # Only define position, size automatic based on widget sizes,
        # instead of format(pixels_x, round(pixels_y*1.05), x_cordinate, y_cordinate))
        root.geometry("+{}+{}".format(x_cordinate, y_cordinate))
        pb          = ttk.Progressbar(root, orient='horizontal', length=pixels_x, mode='determinate')
        pb.pack()

        if platform == "linux" or platform == "linux2":
            root.overrideredirect(True)
            root.wait_visibility(root)
            root.wm_attributes("-alpha", 0.5)
        elif platform == "darwin":
            root.overrideredirect(True)
            # Make the root window always on top
            root.wm_attributes("-topmost", True)
            # Turn off the window shadow
            root.wm_attributes("-transparent", True)
            # Set the root window background color to a transparent color
            root.config(bg='systemTransparent')
            # Store the PhotoImage to prevent early garbage collection
            root.image = img
            # Display the image on a label
            label = tk.Label(root, image=root.image)
            # Set the label background color to a transparent color
            label.config(bg='systemTransparent')
            label.pack()
        elif platform == "win32":
            root.image = img
            label = tk.Label(root, image=root.image, bg='white')
            root.overrideredirect(True)
            root.lift()
            root.wm_attributes("-topmost", True)
            root.wm_attributes("-disabled", True)
            root.wm_attributes("-transparentcolor", "white")
            label.pack()

        return root, pb
    

    def report_progress(self, target, pb, increment):
        target.attributes('-topmost', True)
        pb.step(increment)
        target.update()


#Prepare userland =========================================================
from PyQt5                                  import QtCore, QtWidgets, QtGui
from pyqtgraph                              import PlotWidget
import pyqtgraph                            as pg
from numpy import max, abs, array, zeros, reshape
import os


class GUIWidgets():

    def __init__(self, mainwindow, processing, parameter):

        super(GUIWidgets, self).__init__() # Init Processing class first

        # Load parameters
        # -----------------------------------------------------------------
        self.displ_chans    = [i for i in range(parameter.max_chans) if parameter.selected_chans[i]]
        self.numchans       = len(self.displ_chans)
        self.count          = 0
        self.yrange         = parameter.yrange
        self.bpass          = parameter.bpass
        self.denv           = parameter.dispenv

        # Defaults
        self.streamed_data_type = parameter.streamed_data_type
        self.streaming      = True
        self.envelope       = False
        if self.streamed_data_type == "EEG":
            self.notch      = parameter.notch
            self.bSB        = processing.b_notch
            self.aSB        = processing.a_notch
            self.bPB        = processing.b_wholerange
            self.aPB        = processing.a_wholerange
        else:
            self.notch      = 0
            self.bSB        = array([None, None])
            self.aSB        = array([None, None])
            self.bPB        = processing.b_detrend
            self.aPB        = processing.a_detrend
        self.lighttheme     = QtGui.QGuiApplication.palette()
        self.darktheme      = self.define_darktheme()
        self.darkmode       = parameter.darkmode
        self.mainwindow     = mainwindow

        self.proc           = processing

        self.fps_update_timestamp= 1000 # Setting this to 0 will throw ZeroDivisionError
        self.plot_updates   = 0 # Used as frames per second


    def initiate_theme(self):
        if self.darkmode:
            self.apply_dark_theme()
        elif not self.darkmode:
            self.apply_light_theme()
        

    def fg_vert_range(self):
        # -----------------------------------------------------------------
        # Vert. range (uV)
        # rbt1 (Auto) rdbt2 (100) rdbt3 (200) rdbt4 (500) rdbt5 (1000)
        # -----------------------------------------------------------------

        vertranges          = [0, 200, 500, 1000, "custom"] # 0 is auto scaling

        self.vert_range     = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        horilayout          = QtWidgets.QHBoxLayout()
        title               = QtWidgets.QLabel('Vertical range (uV)')
        rbtn1               = QtWidgets.QRadioButton('Auto')
        rbtn2               = QtWidgets.QRadioButton(str(vertranges[1]))
        rbtn3               = QtWidgets.QRadioButton(str(vertranges[2]))
        rbtn4               = QtWidgets.QRadioButton(str(vertranges[3]))
        rbtn5               = QtWidgets.QRadioButton(str(vertranges[4]))
        self.range_input    = QtWidgets.QLineEdit(str(self.yrange[1]))
        self.range_input.setGeometry(0, 0, 50, self.range_input.geometry().width())
        self.range_input.setFixedWidth(75)
        self.range_input.setDisabled(True)
        
        if self.yrange[1] == vertranges[0]:
            rbtn1.setChecked(True)
        elif self.yrange[1] == vertranges[1]:
            rbtn2.setChecked(True)
        elif self.yrange[1] == vertranges[2]:
            rbtn3.setChecked(True)
        elif self.yrange[1] == vertranges[3]:
            rbtn4.setChecked(True)
        else:
            rbtn5.setChecked(True)
            self.range_input.setEnabled(True)

        rbtn1.clicked.connect(lambda: self.yrange_selection(vertranges[0], title, self.range_input))
        rbtn2.clicked.connect(lambda: self.yrange_selection(vertranges[1], title, self.range_input))
        rbtn3.clicked.connect(lambda: self.yrange_selection(vertranges[2], title, self.range_input))
        rbtn4.clicked.connect(lambda: self.yrange_selection(vertranges[3], title, self.range_input))
        rbtn5.clicked.connect(lambda: self.enable_custom_input(self.range_input))
        self.range_input.returnPressed.connect(lambda: self.custom_yrange(
            self.range_input.text(), [rbtn1, rbtn2, rbtn3, rbtn4], title))


        vertlayout.addWidget(title)
        vertlayout.addLayout(horilayout)
        horilayout.addWidget(rbtn1)
        horilayout.addWidget(rbtn2)
        horilayout.addWidget(rbtn3)
        horilayout.addWidget(rbtn4)
        horilayout.addWidget(rbtn5)
        horilayout.addWidget(self.range_input)
        self.vert_range.setLayout(vertlayout)

        return self.vert_range
    

    def fg_notch_filter(self):
        # -----------------------------------------------------------------
        # Notch filter
        # rbt1 (50 Hz) rdbt2 (60 Hz) rdbt3 (Off)
        # -----------------------------------------------------------------
        self.notch_filter   = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        horilayout          = QtWidgets.QHBoxLayout()
        title               = QtWidgets.QLabel('Notch filter (Hz)')
        rbtn1               = QtWidgets.QRadioButton('50')
        rbtn2               = QtWidgets.QRadioButton('60')
        rbtn3               = QtWidgets.QRadioButton('Off')

        if self.streamed_data_type != "EEG":
            rbtn1.setChecked(False)
            rbtn1.setDisabled(True)
            rbtn1.setCheckable(False)
            rbtn2.setChecked(False)
            rbtn2.setDisabled(True)
            rbtn2.setCheckable(False)
            rbtn3.setChecked(True)
            self.notch == 0
        else:
            if self.notch == 50:
                rbtn1.setChecked(True)
            elif self.notch == 60:
                rbtn2.setChecked(True)
            elif self.notch == 0:
                rbtn3.setChecked(True)

        

        rbtn1.clicked.connect(lambda: self.filt_noise(50))
        rbtn2.clicked.connect(lambda: self.filt_noise(60))
        rbtn3.clicked.connect(lambda: self.filt_noise(0))

        vertlayout.addWidget(title)
        vertlayout.addLayout(horilayout)
        horilayout.addWidget(rbtn1)
        horilayout.addWidget(rbtn2)
        horilayout.addWidget(rbtn3)
        self.notch_filter.setLayout(vertlayout)

        return self.notch_filter
    

    def fg_bandpass_filter(self):
        # -----------------------------------------------------------------
        # Bandpass (Hz)
        # rbt1 (0.5 - 45) rdbt2 (1 - 30) rdbt3 (4 - 8)
        # -----------------------------------------------------------------
        self.bandpass_filter= QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        horilayout          = QtWidgets.QHBoxLayout()
        title               = QtWidgets.QLabel('Bandpass (Hz)')
        rbtn1               = QtWidgets.QRadioButton('Raw')
        rbtn2               = QtWidgets.QRadioButton('Detrend')
        rbtn3               = QtWidgets.QRadioButton('0.5 - 45')
        rbtn4               = QtWidgets.QRadioButton('1 - 30')
        rbtn5               = QtWidgets.QRadioButton('4 - 8')
        rbtn6               = QtWidgets.QRadioButton('0.5 - 6')

        if self.streamed_data_type == "EEG":
            if self.bpass == -1:
                rbtn1.setChecked(True)
            elif self.bpass == 0:
                rbtn2.setChecked(True)
            elif self.bpass == 1:
                rbtn3.setChecked(True)
            elif self.bpass == 2:
                rbtn4.setChecked(True)
            elif self.bpass == 3:
                rbtn5.setChecked(True)
            elif self.bpass == 4:
                rbtn6.setChecked(True)
        else:
            if self.bpass == -1:
                rbtn1.setChecked(True)
            elif self.bpass == 0:
                rbtn2.setChecked(True)
            elif self.bpass == 4:
                rbtn6.setChecked(True)
            rbtn3.setChecked(False)
            rbtn3.setDisabled(True)
            rbtn3.setCheckable(False)
            rbtn4.setChecked(False)
            rbtn4.setDisabled(True)
            rbtn4.setCheckable(False)
            rbtn5.setChecked(False)
            rbtn5.setDisabled(True)
            rbtn5.setCheckable(False)

        rbtn1.clicked.connect(lambda: self.filt_bandpass(-1))
        rbtn2.clicked.connect(lambda: self.filt_bandpass(0))
        rbtn3.clicked.connect(lambda: self.filt_bandpass(1))
        rbtn4.clicked.connect(lambda: self.filt_bandpass(2))
        rbtn5.clicked.connect(lambda: self.filt_bandpass(3))
        rbtn6.clicked.connect(lambda: self.filt_bandpass(4))

        vertlayout.addWidget(title)
        vertlayout.addLayout(horilayout)
        horilayout.addWidget(rbtn1)
        horilayout.addWidget(rbtn2)
        horilayout.addWidget(rbtn3)
        horilayout.addWidget(rbtn4)
        horilayout.addWidget(rbtn5)
        horilayout.addWidget(rbtn6)
        self.bandpass_filter.setLayout(vertlayout)

        return self.bandpass_filter
    

    def fg_envelope(self):
        # -----------------------------------------------------------------
        # Display envelope
        # rbt1 (Off) rdbt2 (On)
        # -----------------------------------------------------------------
        self.envel          = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        horilayout          = QtWidgets.QHBoxLayout()
        title               = QtWidgets.QLabel('Display envelope')
        rbtn1               = QtWidgets.QRadioButton('Off')
        rbtn2               = QtWidgets.QRadioButton('On')

        if self.denv == False:
            rbtn1.setChecked(True)
        elif self.denv == True:
            rbtn2.setChecked(True)

        rbtn1.clicked.connect(lambda: self.disp_envelope(False))
        rbtn2.clicked.connect(lambda: self.disp_envelope(True))

        vertlayout.addWidget(title)
        vertlayout.addLayout(horilayout)
        horilayout.addWidget(rbtn1)
        horilayout.addWidget(rbtn2)
        self.envel.setLayout(vertlayout)

        return self.envel
    

    def fg_stream_button(self):
        # -----------------------------------------------------------------
        # Data stream
        # Dynamic button (Start/Stop (default))
        # -----------------------------------------------------------------
        self.stream         = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        title               = QtWidgets.QLabel('Data stream')
        self.streambtn      = QtWidgets.QPushButton(text="Started")

        self.streambtn.clicked.connect(self.streamstate)

        vertlayout.addWidget(title)
        vertlayout.addWidget(self.streambtn)
        self.stream.setLayout(vertlayout)
        
        return self.stream
    

    def fg_theme_button(self):
        self.theme          = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        title               = QtWidgets.QLabel('Current theme')

        if self.darkmode:
            self.themebtn   = QtWidgets.QPushButton(text="Dark")
        elif not self.darkmode:
            self.themebtn   = QtWidgets.QPushButton(text="Light")

        self.themebtn.clicked.connect(self.themestate)

        vertlayout.addWidget(title)
        vertlayout.addWidget(self.themebtn)
        self.theme.setLayout(vertlayout)
        
        return self.theme


    def fg_static_info(self, udp_ip, udp_port):
        self.headless       = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        title               = QtWidgets.QLabel("""
You are running the GUI headlessly (no GUI displayed).
The raw data is available at {}:{}""".format(udp_ip, udp_port))
        vertlayout.addWidget(title)
        self.headless.setLayout(vertlayout)

        return self.headless
    

    def fg_signal_stream(self, num_samples, s_down, displ_chans, sampling_rate):
        self.signal         = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()

        self.x              = list(range(-num_samples, 0, s_down))
        self.x              = [x/sampling_rate for x in self.x]
        self.xplot          = array(self.x) # Non-sense action to avoid pointers (force creation of other id())
        self.y              = zeros((len(displ_chans), int(num_samples/s_down)), dtype=float)
        self.data_line      = {}
        self.signalgraph    = {}
        self.penstyle       = {}

        for iChan in range(len(displ_chans)):

            self.signalgraph[iChan] = PlotWidget()

            # Aesthetics
            self.signalgraph[iChan].setBackground((255,255,255, 0))
            self.signalgraph[iChan].setLabel('left', 'A (uV)')
            if iChan == len(displ_chans) -1:
                self.signalgraph[iChan].setLabel('bottom', '')
            else:
                self.signalgraph[iChan].setLabel('bottom', '')
                # self.signalgraph[iChan].getAxis('bottom').setTicks([])
            self.signalgraph[iChan].showGrid(x=False, y=True)
            self.penstyle[iChan] = pg.mkPen(color=(49,130,189), width=2)
            
            # Decorate plot
            legend = self.signalgraph[iChan].addLegend(offset=(5, 0))
            self.signalgraph[iChan].setAntialiasing(True)  # Huge performance gain and 
                                                # necessary to keep up with 
                                                # the sampling rate
            self.signalgraph[iChan].setYRange(self.yrange[0], self.yrange[1])
            self.signalgraph[iChan].setRange(
                yRange=(self.yrange[0], self.yrange[1]), 
                padding=0, disableAutoRange=True)
            self.signalgraph[iChan].setMinimumSize(200, 50)
            
            # Set signal lines
            self.data_line[iChan] = self.signalgraph[iChan].plot(
                self.xplot, self.y[iChan,:],
                name='Ch. {}'.format(str(displ_chans[iChan]+1)),
                pen=self.penstyle[iChan])

            # Disable interactivity
            self.signalgraph[iChan].setMouseEnabled(x=False, y=False)
            self.signalgraph[iChan].hideButtons()
            self.signalgraph[iChan].getPlotItem().setMenuEnabled(False)
            legend.mouseDragEvent = lambda *args, **kwargs: None
            legend.hoverEvent = lambda *args, **kwargs: None

            # Fix sizes in order to avoid filling of white spaces
            # self.signalgraph[iChan].getViewBox().size()
            # self.signalgraph[iChan].getAxis('bottom').setWidth(
            #     self.signalgraph[iChan].getAxis('bottom').size().width())

            vertlayout.addWidget(self.signalgraph[iChan])

        self.signal.setLayout(vertlayout)

        return self.signal
    

    # def frontend_thread(self, s_down, left_edge, sampling_rate,
    #                        idx_retain, max_chans, displ_chans,
    #                        shared_buffer, shared_timestamp, gui_running):
        
    #     while gui_running.value == 1:
    #         self.update_signal_plot(s_down, left_edge, sampling_rate,
    #                        idx_retain, max_chans, displ_chans,
    #                        shared_buffer, shared_timestamp)


    def update_fps(self):
        self.fps_info.setText(" FPS \n{}".format(self.plot_updates))
        # White spaces to avoid widget resizing when number of digits change


    def update_signal_plot(self, s_down, left_edge, sampling_rate,
                           idx_retain, max_chans, displ_chans,
                           shared_buffer, shared_timestamp):
        
        self.plot_updates += 1
                
        self.count = self.count + 1
        if self.count < s_down:
            return

        # Reconstruct buffer
        # -----------------------------------------------------------------
        # sharedbuffer is a 1D array where channels are concatenated. We
        # split them up into 2D arrays again
        buffer              = self.rebuild_buffer(shared_buffer[:], max_chans)
                
        # Filter buffer signal and send filtered data to plotting funcs
        # -------------------------------------------------------------
        processed_buffer    = self.prepare_buffer(buffer, 
            self.bSB, self.aSB, self.bPB, self.aPB)
        processed_buffer    = processed_buffer[:, left_edge:]

        if self.envelope == True:
            processed_buffer = self.extract_envelope(processed_buffer)

        x_current           = shared_timestamp.value / 1000
        x_first             = x_current - len(self.x) * s_down / sampling_rate
        self.x = [s_down / sampling_rate * i + x_first for i in range(len(self.x))] 
        
        # self.x              = list(
        #     range(
        #         round(x_first*1000),
        #         round(x_current*1000),
        #         int(1000/sampling_rate*s_down)))
        # self.x              = [i / 1000 for i in self.x]


        if self.streaming == True:
            self.xplot      = self.x

        
        # Downsample buffer
        down_buffer = processed_buffer[:, idx_retain]

        # Set vertical range
        if self.streaming == False:
            vscale = self.signalgraph[0].getAxis("left").range
        elif self.yrange[1] == 0:
            v_buffer    = reshape(processed_buffer, processed_buffer.size)
            vscale = [
                -max([abs(min(v_buffer)), abs(max(v_buffer))]),
                max([abs(min(v_buffer)), abs(max(v_buffer))])]
        else:
            vscale = list(self.yrange)

        
        # Correct units by 10^3
        if vscale[1] >= 1000000:
            vscale              = [x / 1000000 for x in vscale]
            down_buffer         = down_buffer / 1000000
            amp_label           = 'A (V)'
        elif vscale[1] >= 1000:
            vscale              = [x / 1000 for x in vscale]
            down_buffer         = down_buffer / 1000
            amp_label           = 'A (mV)'
        else:
            amp_label           = 'A (uV)'
            
        

        # Update plots for every channel
        # -------------------------------------------------------------
        for iChan in range(len(displ_chans)):

            if self.streaming == True:
                self.y[iChan,:] = down_buffer[displ_chans[iChan], :]
            
            self.data_line[iChan].setData(self.x, self.y[iChan])  # Update the data

            if self.envelope == True:
                self.signalgraph[iChan].setYRange(0, vscale[1], padding=0)
            else:
                self.signalgraph[iChan].setYRange(vscale[0], vscale[1], padding=0)

            self.signalgraph[iChan].setLabel('left', amp_label)

        self.count          = 0

        if self.fps_update_timestamp + 1000 < shared_timestamp.value:
            self.update_fps()
            self.plot_updates = 0
            self.fps_update_timestamp = shared_timestamp.value


    def rebuild_buffer(self, buffer_in, num_channels):
        return reshape(buffer_in, (num_channels, int(len(buffer_in)/num_channels)))

    
    def filt_noise(self, choice):
        choice = int(choice)
        if choice == 50:
            print('Enabled 50 Hz stopband filter')
            self.bSB    = self.proc.b_notch
            self.aSB    = self.proc.a_notch
        elif choice == 60:
            print('Enabled 60 Hz stopband filter')
            self.bSB    = self.proc.b_notch60
            self.aSB    = self.proc.a_notch60
        elif choice == 0:
            print('Notch filter disabled')
            self.bSB    = array([None, None]) # Avoiding bool not iterable
            self.aSB    = array([None, None])


    def filt_bandpass(self, choice):
        choice = int(choice)
        if choice == -1:
            print('Displaying raw signal')
            self.bPB        = array([None, None])
            self.aPB        = array([None, None])
        elif choice == 0:
            print('Highpass filter from 0.1 Hz')
            self.bPB        = self.proc.b_detrend
            self.aPB        = self.proc.a_detrend
        elif choice == 1:
            print('Bandpass filter between 0.5 and 45 Hz')
            self.bPB        = self.proc.b_wholerange
            self.aPB        = self.proc.a_wholerange
        elif choice == 2:
            print('Bandpass filter between 1 and 30 Hz')
            self.bPB        = self.proc.b_sleep
            self.aPB        = self.proc.a_sleep
        elif choice == 3:
            print('Bandpass filter between 4 and 8 Hz')
            self.bPB        = self.proc.b_theta
            self.aPB        = self.proc.a_theta
        elif choice == 4:
            print('Bandpass filter between 0.5 and 6 Hz')
            self.bPB        = self.proc.b_slow
            self.aPB        = self.proc.a_slow


    def yrange_selection(self, choice, title, custom_input):
        choice              = int(choice)
        if choice == 100:
            print('Vertical range set to -100 uV to +100 uV')
            self.yrange     = (-100, 100)
        elif choice == 200:
            print('Vertical range set to -200 uV to +200 uV')
            self.yrange     = (-200, 200)
        elif choice == 500:
            print('Vertical range set to -500 uV to +500 uV')
            self.yrange     = (-500, 500)
        elif choice == 1000:
            print('Vertical range set to -1000 uV to +1000 uV')
            self.yrange     = (-1000, 1000)
        elif choice == 0:
            print('Vertical range set relative to signal')
            self.yrange     = (-0, 0)
        title.setText('Vertical range (uV)')
        custom_input.setText(str(self.yrange[1]))
        custom_input.setDisabled(True)


    def enable_custom_input(self, custom_input):
        custom_input.setEnabled(True)


    def custom_yrange(self, choice, radio_buttons, title):

        try: 
            choice              = abs(int(choice))
            self.yrange         = (-choice, choice)
            title.setText('Vertical range (uV)')
            print('Vertical range set to -{} uV to +{} uV'.format(choice, choice))
        except:
            title.setText("Set an integer number")
            print("You have to chose a positive integer number")


    def disp_envelope(self, choice):
        if choice == True:
            print('Enabled envelope displaying')
            self.envelope = True
        elif choice == False:
            print('Disabled envelope displaying')
            self.envelope = False


    def streamstate(self):
        if not self.streaming:
            self.streambtn.setText('Started')
            self.streaming = True
        elif self.streaming:
            self.streambtn.setText('Stopped')
            self.streaming = False


    def themestate(self):

        if self.darkmode:
            self.themebtn.setText('Light')
            self.apply_light_theme()
            self.darkmode = False
            self.save_parameters()
            print('Enabled light theme')
        elif not self.darkmode:
            self.themebtn.setText('Dark')
            self.apply_dark_theme()
            self.darkmode = True
            self.save_parameters()
            print('Enabled dark theme')


    def display_fps(self):
        fps_display         = QtWidgets.QWidget()
        vertlayout          = QtWidgets.QVBoxLayout()
        self.fps_info       = QtWidgets.QLabel(" FPS \n{}".format(self.plot_updates))
        self.fps_info.setAlignment(QtCore.Qt.AlignCenter)
        vertlayout.addWidget(self.fps_info)
        fps_display.setLayout(vertlayout)

        return fps_display


    def save_parameters(self):

        end_line        = "\n"
        conf_file       = './settings.cfg'

        if not os.path.exists(conf_file): # File generation

            with open(conf_file, 'w') as f:
                f.write("".join(["Darkmode=", str(self.darkmode), end_line]))

        else:

            with open(conf_file, 'r') as f:
            
                settings                        = f.readlines()
                
                for i, setting in enumerate(settings): # Update values
                    if 'Darkmode' in setting:
                        settings[i]             = "".join(["Darkmode=", str(self.darkmode), end_line])

                new_settings = []
                if len([s for s in settings if "Darkmode" in s]) == 0:
                    new_settings.append("".join(["Darkmode=", str(self.darkmode), end_line]))

            with open(conf_file, 'w') as f:
                f.write("".join(settings + new_settings))


    def define_darktheme(self):
        darktheme = QtGui.QPalette()
        darktheme.setColor(QtGui.QPalette.Window, QtGui.QColor(53, 53, 53))
        darktheme.setColor(QtGui.QPalette.WindowText, QtCore.Qt.white)
        darktheme.setColor(QtGui.QPalette.Base, QtGui.QColor(25, 25, 25))
        darktheme.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(53, 53, 53))
        darktheme.setColor(QtGui.QPalette.ToolTipBase, QtCore.Qt.black)
        darktheme.setColor(QtGui.QPalette.ToolTipText, QtCore.Qt.white)
        darktheme.setColor(QtGui.QPalette.Text, QtCore.Qt.white)
        darktheme.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
        darktheme.setColor(QtGui.QPalette.ButtonText, QtCore.Qt.white)
        darktheme.setColor(QtGui.QPalette.BrightText, QtCore.Qt.red)
        darktheme.setColor(QtGui.QPalette.Link, QtGui.QColor(42, 130, 218))
        darktheme.setColor(QtGui.QPalette.Highlight, QtGui.QColor(42, 130, 218))
        darktheme.setColor(QtGui.QPalette.HighlightedText, QtCore.Qt.black)

        return darktheme


    def apply_light_theme(self):
        self.signal.setPalette(self.lighttheme)
        self.theme.setPalette(self.lighttheme)
        self.stream.setPalette(self.lighttheme)
        self.bandpass_filter.setPalette(self.lighttheme)
        self.notch_filter.setPalette(self.lighttheme)
        self.envel.setPalette(self.lighttheme)
        self.vert_range.setPalette(self.lighttheme)
        self.streambtn.setPalette(self.lighttheme)
        self.themebtn.setPalette(self.lighttheme)
        self.headless.setPalette(self.lighttheme)
        self.range_input.setStyleSheet(
            'QLineEdit {background-color: rgb(255,255,255); color: rgb(0,0,0);}')
        self.streambtn.setStyleSheet(
            'QPushButton {background-color: rgb(222,235,247); border-radius: 10px;padding: 6px;color: #08519c;}')
        self.themebtn.setStyleSheet(
            'QPushButton {background-color: rgb(222,235,247); border-radius: 10px;padding: 6px;color: #08519c;}')
        self.mainwindow.setPalette(self.lighttheme)

        for iChan in range(self.numchans):
            self.signalgraph[iChan].setBackground((247,247,247,0))
            self.penstyle[iChan] = pg.mkPen(color=(49,130,189), width=2)


    def apply_dark_theme(self):        
        self.signal.setPalette(self.darktheme)
        self.theme.setPalette(self.darktheme)
        self.stream.setPalette(self.darktheme)
        self.bandpass_filter.setPalette(self.darktheme)
        self.notch_filter.setPalette(self.darktheme)
        self.envel.setPalette(self.darktheme)
        self.vert_range.setPalette(self.darktheme)
        self.streambtn.setPalette(self.darktheme)
        self.themebtn.setPalette(self.darktheme)
        self.headless.setPalette(self.darktheme)
        self.range_input.setStyleSheet(
            'QLineEdit {background-color: rgb(53, 53, 53); color: #eff3ff;}')
        self.streambtn.setStyleSheet(
            'QPushButton {background-color: rgb(49,130,189); border-radius: 10px;padding: 6px; color: #eff3ff;}')
        self.themebtn.setStyleSheet(
            'QPushButton {background-color: rgb(49,130,189); border-radius: 10px;padding: 6px; color: #eff3ff;}')
        self.mainwindow.setPalette(self.darktheme)

        for iChan in range(self.numchans):
            self.signalgraph[iChan].setBackground((53, 53, 53,0))
            self.penstyle[iChan] = pg.mkPen(color=(222,235,247), width=2)


try:
    from compatible_boards import COMPATIBLE_BOARDS
except:
    from .compatible_boards import COMPATIBLE_BOARDS


class BoardAgnosticUtils():
        
    def __init__(self, parameter):

        # We can not just assign all parameters to for example self.params
        # since Python's Process() does not allow to parse several of the
        # stored objects in "parameters"
        self.streaming_parameter = {
            "baud_rate":    parameter.baud_rate,
            "time_out":     parameter.time_out,
            "port":         parameter.port,
            "start_code":   parameter.start_code, # Start code that boards reads as sign to startt sampling (int)
            "max_chans":    parameter.max_chans, # Specifies how many channels the incoming signals has (int)
            "buffer_length":parameter.buffer_length, # Length of constructed signal buffer (int)
            "buffer_add":   parameter.buffer_add, # Length of padding to avoid filter artifacts  (int)
            "sample_rate":  parameter.sample_rate, # Sampling rate (Hz, int)
            "PGA":          parameter.PGA, # Programmable gain (int)
            "saving_interval": parameter.saving_interval, # how often the data gets written to disk (seconds, int)
            "udp_ip":       parameter.udp_ip, # IP the data shall be forwarded to
            "udp_port":     parameter.udp_port, # Port of signal forwarding
            "set_customsession": parameter.set_customsession, # Whether or not to have a custom output file name (bool)
            "sessionName":  parameter.sessionName, # Custom file name for output (str)
            "streamed_data_type": parameter.streamed_data_type, # Data type to stream (str)
        }


    def get_board_specific_utils(self, board):
        """Returns the board_specific functions that allows us to read and 
        process incoming signals

        Args:
            board (str): Unique name of board. Defined in parameters.py

        Returns:
            (object): Class object containing all necessary functions to
                work with given board.
        """
        for _, board_name in enumerate(list(COMPATIBLE_BOARDS.keys())):
            if board == board_name:
                print(f"Will use {COMPATIBLE_BOARDS[board_name][3].__qualname__} to communicate with board")
                return COMPATIBLE_BOARDS[board_name][3]
        raise Exception("Could not get board-specific functions collection")


    def headless_sampling(self):
        pass # This function only exists in order to make timer call something

    
         



from json import dumps
from time import sleep, perf_counter
from numpy import expand_dims, concatenate, append, array, zeros, reshape
try:
    from io_manager import IOManager
except:
    from .io_manager import IOManager


class SamplingUtilsBioAmpUpsideDownLabs():
    
    def __init__(self):
        pass # Necessary init function because we call attributes of this class in process_samples()

    
    def message_to_samples(self, str_message, s_chans):
        # -----------------------------------------------------------------
        # Extract samples from board
        # Input
        #   str_message (str) Raw message string coming from board
        #   s_chans     (int) How many channels shall the samples be 
        #               assigned to
        #   board code  (int) Defines code pipeline based on board type
        # Default in case of faulty message
        eeg_array       = expand_dims(
            array([0] * s_chans, dtype=float), # Crucial to set float
            axis=1)
        eeg_valid       = False
        #   eeg_array   Numpy array (floats), being channels by samples
        #   eeg_valid   Boolean whether workable data or not
        # -----------------------------------------------------------------
        str_message = str_message.replace("b'", "")
        str_message = str_message.replace("\\r\\n\'", "")

        try:
            eeg_array[0, 0]   = float(str_message)
        except ValueError:
            return eeg_array, eeg_valid
        
        eeg_valid       = True
        return eeg_array, eeg_valid


    def process_samples(self, transmitter, parameter, shared_buffer,
                     shared_timestamp, gui_running):

        io_manager = IOManager(parameter)
        receiver = io_manager.set_up_serial_entry_point(
            parameter["baud_rate"], parameter["time_out"], parameter["port"])
        print('Ordered board to send data via USB. Switching mode ...')

        with receiver as r:

            # Open communication ----------------------------------------------
            sleep(1)

            # Prealloate values of loop ---------------------------------------
            start_time          = int(perf_counter() * 1000)
            time_stamp_now      = int(perf_counter() * 1000) # Do NOT copy from start_time (will generate pointer)
            time_reset          = int(perf_counter() * 1000) # Do NOT copy from start_time (will generate pointer)
            sample_count        = int(0)
            
            buffer              = zeros((parameter.max_chans, 
                (parameter.buffer_length + parameter.buffer_add) * parameter.sample_rate))
            time_stamps         = zeros(parameter.buffer_length * parameter.sample_rate, dtype=int)
            pga                 = parameter.PGA
            s_chans             = parameter.max_chans
            sampling_rate       = parameter.sample_rate
            saving_interval     = parameter.saving_interval * parameter.sample_rate
            update_buffer       = zeros((buffer.shape[0], buffer.shape[1]+1))
            update_times        = zeros((buffer.shape[1]+1), dtype=int)

            # Preallocate json relay message and relay connection
            relay_array = IOManager.build_standard_relay_message()
            udp_ip              = parameter.udp_ip
            udp_port            = parameter.udp_port

            self.empty_message_queue(r)
                
            while gui_running.value == 1:
                
                raw_message             = str(r.readline())

                buffer_in, valid_eeg    = self.message_to_samples(
                    raw_message, s_chans)

                if not valid_eeg:
                    # TODO: Implement an interpolation system
                    continue

                # Current timestamp -------------------------------------------
                time_stamp_now          = int(perf_counter() * 1000) - start_time
                # This will generate unchanged time_stamps for all samples of 
                # the incoming bufer (= 10 in case of bluetooth), but that is
                # not a problem

                sample_count        = sample_count + 1
                
                sample              = buffer_in[:, 0]

                # Convert binary to voltage values
                for iS in range(s_chans):
                    relay_array["".join(["c", str(iS+1)])] = str(sample[iS])

                update_buffer       = concatenate((buffer, expand_dims(sample, 1)), axis=1)
                update_times        = append(time_stamps, time_stamp_now)

                # Build new buffer and timestamp arrays
                buffer              = update_buffer[:, 1:]
                time_stamps         = update_times[1:]

                # Make data available for downstream programs
                transmitter.sendto(bytes(dumps(relay_array), "utf-8"), (udp_ip, udp_port))

                # Update shared memory allocations for frontend
                shared_buffer[:] = reshape(buffer, buffer.size) # Has left edge for filtering
                shared_timestamp.value = time_stamp_now

                # Write out samples to file -----------------------------------
                if sample_count == saving_interval:

                    IOManager.calc_sample_rate(time_stamp_now, time_reset, 
                                        sampling_rate, time_stamps)

                    IOManager.master_write_data(buffer, time_stamps, 
                        saving_interval)
                    sample_count        = 0
                    time_reset          = time_stamp_now

            r.write(bytes(str(0), 'utf-8')) # Set board into standby
            sleep(1)
            return


"""Use this as a template of how to make your board compatible with the 
NeuriGUI.
Requirement is that you maintain everything inside the template and only 
add your own steps to the existing ones. This is because the GUI has strict
methods of importing classes and fucntions as well as sharing variables 
between frontend and backend.
"""


from json import dumps
from time import sleep, perf_counter
from numpy import expand_dims, concatenate, append, array, zeros, reshape
try:
    from io_manager import IOManager
except:
    from .io_manager import IOManager
    

class SamplingUtilsTemplate():

    def __init__(self):
        pass # Necessary init function because we call attributes of this class in process_samples()
    

    def message_to_samples(self, str_message, s_chans):
        # -----------------------------------------------------------------
        # Extract samples from board
        # Input
        #   str_message (str) Raw message string coming from board
        #   s_chans     (int) How many channels shall the samples be 
        #               assigned to
        # Default in case of faulty message
        eeg_array       = expand_dims(
            array([0] * s_chans, dtype=float), # Crucial to set float
            axis=1)
        eeg_valid       = False
        #   eeg_array   Numpy array (floats), being channels by samples
        #       (Has to be 2D even if only one channel present [samples x 1])
        #   eeg_valid   Boolean whether workable data or not
        # -----------------------------------------------------------------
        
        # TODO: Extract channels' samples from str_message and parse into
        # eeg_array
        
        return eeg_array, eeg_valid


    def process_samples(self, transmitter, parameter, shared_buffer,
                     shared_timestamp, gui_running):
        """
        parameter = {
            "baud_rate":    parameter.baud_rate,
            "time_out":     parameter.time_out,
            "port":         parameter.port,
            "start_code":   parameter.start_code, # Start code that boards reads as sign to startt sampling (int)
            "max_chans":    parameter.max_chans, # Specifies how many channels the incoming signals has (int)
            "buffer_length":parameter.buffer_length, # Length of constructed signal buffer (int)
            "buffer_add":   parameter.buffer_add, # Length of padding to avoid filter artifacts  (int)
            "sample_rate":  parameter.sample_rate, # Sampling rate (Hz, int)
            "PGA":          parameter.PGA, # Programmable gain (int)
            "saving_interval": parameter.saving_interval, # how often the data gets written to disk (seconds, int)
            "udp_ip":       parameter.udp_ip, # IP the data shall be forwarded to
            "udp_port":     parameter.udp_port, # Port of signal forwarding
            "set_customsession": parameter.set_customsession, # Whether or not to have a custom output file name (bool)
            "sessionName":  parameter.sessionName, # Custom file name for output (str)
        }"""

        io_manager = IOManager(parameter)
        receiver = io_manager.set_up_serial_entry_point(
            parameter["baud_rate"], parameter["time_out"], parameter["port"])

        with receiver as r:

            # Open communication ----------------------------------------------
            sleep(1) # Allow for connection to stabilize

            # TODO: Send start_code to board if needed in order to 
            # initialize sample transmission from board to GUI

            # Prealloate values of loop ---------------------------------------
            start_time          = int(perf_counter() * 1000)
            time_stamp_now      = int(perf_counter() * 1000) # Do NOT copy from start_time (will generate pointer)
            time_reset          = int(perf_counter() * 1000) # Do NOT copy from start_time (will generate pointer)
            sample_count        = int(0)
            
            buffer              = zeros((parameter["max_chans"],
                (parameter["buffer_length"] + parameter["buffer_add"]) * parameter["sample_rate"]))
            time_stamps         = zeros(parameter["buffer_length"] * parameter["sample_rate"], dtype=int)
            pga                 = parameter["PGA"]
            s_chans             = parameter["max_chans"]
            sampling_rate       = parameter["sample_rate"]
            saving_interval     = parameter["saving_interval"] * parameter["sample_rate"]
            update_buffer       = zeros((buffer.shape[0], buffer.shape[1]+1))
            update_times        = zeros((buffer.shape[1]+1), dtype=int)

            # Preallocate json relay message and relay connection
            relay_array         = io_manager.build_standard_relay_message(parameter["max_chans"])
            udp_ip              = parameter["udp_ip"]
            udp_port            = parameter["udp_port"]
                
            while gui_running.value == 1:
                
                # =========================================================
                # ========= B O A R D   S P E C I F I C   P A R T =========
                # =========================================================
                # TODO: Adapt recevier function to read incoming data from board
                raw_message             = str(r.readline())

                # TODO: Adapt message_to_samples() in order to convert
                # raw_message string into numpy array of samples
                buffer_in, valid_eeg    = self.message_to_samples(raw_message, s_chans)

                if not valid_eeg:
                    # TODO: Implement an interpolation system
                    continue                
                # =========================================================
                # =========================================================
                # =========================================================

                # Standard set of procedures from here on: You should not
                # require any additional changes from here

                # Current timestamp -------------------------------------------
                time_stamp_now          = int(perf_counter() * 1000) - start_time
                # This will generate unchanged time_stamps for all samples of 
                # the incoming buffer (= 10 in case of bluetooth), but that is
                # not a problem

                sample_count        = sample_count + 1
                
                sample              = buffer_in[:, 0]

                for iSample in range(s_chans):
                    relay_array["".join(["c", str(iSample+1)])] = str(sample[iSample])

                update_buffer       = concatenate((buffer, expand_dims(sample, 1)), axis=1)
                update_times        = append(time_stamps, time_stamp_now)

                # Build new buffer and timestamp arrays
                buffer              = update_buffer[:, 1:]
                time_stamps         = update_times[1:]

                # Make data available for downstream programs
                transmitter.sendto(bytes(dumps(relay_array), "utf-8"), (udp_ip, udp_port))

                # Update shared memory allocations for frontend
                shared_buffer[:] = reshape(buffer, buffer.size) # Has left edge for filtering
                shared_timestamp.value = time_stamp_now

                # Write out samples to file -----------------------------------
                if sample_count == saving_interval:

                    io_manager.calc_sample_rate(time_stamp_now, time_reset, 
                                        sampling_rate, time_stamps)

                    io_manager.master_write_data(buffer, time_stamps, 
                        saving_interval)
                    sample_count        = 0
                    time_reset          = time_stamp_now

            r.write(bytes(str(0), 'utf-8')) # Set board into standby
            sleep(1)
            return


import serial
import socket
from threading                          import Thread
from datetime                           import datetime
from pylsl                              import StreamInlet
try:
    from pylsl                          import resolve_stream
except ImportError:
    from pylsl.resolve                  import resolve_stream


class IOManager():

    def __init__(self, parameter):
        #Output
        t0              = str(datetime.now())
        t0              = t0.replace(':', '_')
        if parameter["set_customsession"]:
            file_name   = parameter["sessionName"] + '.txt'
        else:
            file_name   = 'Neuri_' + t0 + '.txt'

        # Prepare data output
        self.output_file= file_name

        with open(self.output_file, 'w', encoding= "utf_8") as file:
            file.write("Session Started\n")
            file.close() # Important for data to get written


    def set_up_serial_entry_point(self, baud_rate, time_out, port):
        # =================================================================
        # Connects to device, preferrably by USB
        # INPUT
        #   baud_rate       = baud rate of the port (scalar)
        #   time_out        = how long a message should be waited for
        #                     default = None (infinite), scalar or None
        #   port            = port to connect to ("COMxy", str)
        # OUTPUT
        #   ser             = Serial connection (object)
        # =================================================================

        # Open communication protocol
        ser            = serial.Serial()
        ser.baudrate   = baud_rate
        ser.timeout    = time_out
        ser.port       = port
        print('Ready to connect to board')
        return ser
    

    def set_up_lsl_entry_point(self):
        streams_eeg = resolve_stream("type", "EEG")
        inlet_eeg = StreamInlet(streams_eeg[0])
        # streams_gyro = resolve_stream("type", "Gyro")
        # inlet_gyro = StreamInlet(streams_gyro[0])
        streams_ppg = resolve_stream("type", "PPG")
        inlet_ppg = StreamInlet(streams_ppg[0])
        return inlet_eeg, inlet_ppg
    

    def set_up_socket_entry_point(self, ip, port):
        receiver_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        receiver_sock.bind((str(ip), int(port)))
        return receiver_sock
    

    def master_write_data(self, eeg_data, time_stamps, saving_interval):
        # =================================================================
        # Input:
        #   eeg_data            2D numpy array [channels x samples] (float)
        #   time_stamps         1D numpy array (float)
        #   saving_interval     Scalar
        # Output:
        #   No output
        # =================================================================
        new_buffer          = eeg_data[:, -saving_interval:]
        new_time_stamps     = time_stamps[-saving_interval:]
            
        save_thread = Thread(target=self.write_data_thread,
            args=(new_buffer, new_time_stamps))
        save_thread.start()


    def calc_sample_rate(self, curr_time, prev_iter_time, sample_rate, time_stamps):

        time_diff           = curr_time - prev_iter_time
        # It took (time_diff) ms to fetch (sample_rate) samples.
        # Calculate actual sampling rate
        actual_sr           = int((sample_rate / (time_diff / 1000)))
        print('%d ms: Writing data (sampling rate = %d Hz)' %
            (time_stamps[-1], actual_sr))


    def write_data_thread(self, eeg_data, time_stamps):
        # =================================================================
        # Input:
        #   eeg_data            2D numpy array [channels x samples] (float)
        #   time_stamps         1D numpy array (float)
        # Output:
        #   No output
        # =================================================================
        # Append the data points to the end of the file
        with open(self.output_file, 'a', encoding= "utf_8") as file:
            buffer_length = time_stamps.shape[0]

            for sample_index in range(buffer_length):
                # Format time stamp
                time_stamp      = time_stamps[sample_index]
                # format eeg data
                eeg_data_points = eeg_data[:,sample_index].tolist()
                eeg_data_points = [str(value) for value in eeg_data_points]
                eeg_data_points = ",".join(eeg_data_points)
                
                # Write line to file
                file.write(f"{time_stamp}, {eeg_data_points} \n")
            file.close() # Important for data to get written


    def build_standard_relay_message(self, s_chans):
        """Build a relay jsonized message string that gets forwarded to any
        third-party application.

        Args:
            s_chans (int): Number of channels (specific for boards)

        Returns relay_array (str):json.format message string with keys for:
            1. timestamps (t)
            2. channels (c)
        """
        relay_array         = {}
        relay_array["t"]    = ''
        for iC in range(s_chans):
            relay_array["".join(["c", str(iC+1)])] = ''
        return relay_array


"""Manual board management
For every board, we collect here the sampling functions and store them 
inside the same class. For this, you need to manage the imports and
dictionnary manually, adding the imported object in fourth position of 
the board's [list].
"""
try:
    from signal_sampling_neuri_v1       import SamplingUtilsNeuriV1Helment
    from signal_sampling_neuri_lolin    import SamplingUtilsNeuriLolinHelment
    from signal_sampling_bioamp         import SamplingUtilsBioAmpUpsideDownLabs
    from signal_sampling_muse_s         import SamplingUtilsMuseSInteraxon
except: # Necessary if running as Python package
    from .signal_sampling_neuri_v1      import SamplingUtilsNeuriV1Helment
    from .signal_sampling_neuri_lolin   import SamplingUtilsNeuriLolinHelment
    from .signal_sampling_bioamp        import SamplingUtilsBioAmpUpsideDownLabs
    from .signal_sampling_muse_s        import SamplingUtilsMuseSInteraxon


# Set None where parameter is not required
COMPATIBLE_BOARDS = {
    # Unique name: [
    #   0: Start code (int or None),
    #   1: Main sampling rate (EEG for most boards) (Hz),
    #   2: Baud rate (int),
    #   3: Board_specific class (object),
    #   4: Number of channels (int),
    #   5: Uses lab streaming layer LSL (bool),
    #   6: Possible to set PGA (bool),
    #   7: Sampling rate of secondary data type applicable, ie PPG (int or None),
    #   8: Amount of channels for secondary data type (ie PPG) (int or None),
    #   9: Default Programmable Gain Amplitude (int or None)
    # ]
    "Neuri V1 by Helment": [
        2,
        200,
        115200,
        SamplingUtilsNeuriV1Helment,
        2,
        False,
        True,
        None,
        None,
        24,
    ],
    "Neuri-Lolin S3-PRO by Helment": [
        2,
        200,
        115200,
        SamplingUtilsNeuriLolinHelment,
        8,
        False,
        True,
        None,
        None,
        24,
    ],
    "BioAmp EXG Pill by Upside Down Labs": [
        None,
        125,
        115200,
        SamplingUtilsBioAmpUpsideDownLabs,
        1,
        False,
        False,
        None,
        None,
        None,
    ],
    "Muse S by InteraXon": [
        None,
        256,
        None,
        SamplingUtilsMuseSInteraxon,
        6,
        True,
        False,
        64,
        3,
        None,
    ]
}


#Prepare userland =========================================================
from multiprocessing                            import Process, Array, Value
from PyQt5                                      import QtCore, QtWidgets
from numpy                                      import zeros
from time                                       import sleep
import sys  # We need sys so that we can pass argv to QApplication

# Necessary step for relative imports when the GUI is run directly in an 
# IDE instead of as module
try:
    from backend.signal_processing              import Processing
    from backend.board_agnostic_utils           import BoardAgnosticUtils
    from backend.parameter_validation           import ParamVal
    from frontend.widgets                       import GUIWidgets
    from frontend.parameters                    import Parameters
except:
    from .backend.signal_processing             import Processing
    from .backend.board_agnostic_utils          import BoardAgnosticUtils
    from .backend.parameter_validation          import ParamVal
    from .frontend.widgets                      import GUIWidgets
    from .frontend.parameters                   import Parameters


class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, *args, **kwargs):

        QtWidgets.QMainWindow.__init__(self)

        # Set parameters
        # -----------------------------------------------------------------
        pm                  = Parameters()
        ParamVal(pm)        # Sanity checks

        # Initialize filter coefficient arrays
        # -----------------------------------------------------------------
        proc                = Processing(pm)

        # Load methods for board handling
        # -----------------------------------------------------------------
        boardutils          = BoardAgnosticUtils(pm)

        # Generate variable exchange variables (shared memory allocations)
        # -----------------------------------------------------------------
        shared_buffer                   = Array(
            'd', zeros(( 
                (pm.buffer_length + pm.buffer_add) * pm.sample_rate * pm.max_chans
                )), lock = True)
        shared_timestamp                = Value('i', 0)
        self.gui_running                = Value('i', 1)

        # Generate separate processes to not slow down sampling by any
        # other executions
        # -----------------------------------------------------------------
        # Here, we can not just parse "pm" as it contains non-pickable tk 
        # objects which will break at self.sampling.start(). We create a 
        # new object with the necessary variables only
        sampling_functions = boardutils.get_board_specific_utils(pm.board)()
        self.sampling    = Process(
            target=sampling_functions.process_samples,
            args=(pm.send_sock,
                  boardutils.streaming_parameter,
                  shared_buffer,
                  shared_timestamp,
                  self.gui_running))

        # Build GUI
        # -----------------------------------------------------------------
        super(MainWindow, self).__init__(*args, **kwargs)
        
        guiwidgets          = GUIWidgets(self, proc, pm)
        # Inherit processing functions that will be used during plot update
        guiwidgets.prepare_buffer   = proc.prepare_buffer
        guiwidgets.extract_envelope = proc.extract_envelope
        
        self.central_widget = QtWidgets.QWidget() # A QWidget to work as Central Widget

        # Without specifying, this just makes the GUI separable vertically
        # and horizontally 
        vertlayout          = QtWidgets.QVBoxLayout() # Vertical Layout
        controlpanel        = QtWidgets.QHBoxLayout() # Horizontal Layout

        # Load frontend elements
        widget_vrange       = guiwidgets.fg_vert_range()
        widget_notch        = guiwidgets.fg_notch_filter()
        widget_bandpass     = guiwidgets.fg_bandpass_filter()
        widget_envelope     = guiwidgets.fg_envelope()
        widget_sbtn         = guiwidgets.fg_stream_button()
        widget_darkmode     = guiwidgets.fg_theme_button()
        widget_fps          = guiwidgets.display_fps()
        widget_signal       = guiwidgets.fg_signal_stream(
            int(pm.sample_rate * pm.buffer_length), pm.s_down,
            [i for i in range(pm.max_chans) if pm.selected_chans[i]],
            pm.sample_rate)
        widget_headless     = guiwidgets.fg_static_info(pm.udp_ip, pm.udp_port)

        guiwidgets.initiate_theme() # Needs to be called after widget 
                                    # elements initiation since hard-coded

        self.setCentralWidget(self.central_widget)
        self.central_widget.setLayout(vertlayout) # Draw elements in main widget
        # auxgui.report_progress(splash, pb, 10)

        self.timer              = QtCore.QTimer()
        self.timer.setInterval(20)  # This can be set to 0 as well for
                                    # fastest plotting calls possible but
                                    # the GUI currently freezes in those
                                    # conditions and I have not figured out
                                    # how to use QThreads accordingly
                                    # (window does not get updated if 
                                    # update plot functions are inside
                                    # QThread)
        self.timer.singleShot   = False

        if pm.run_headless:

            self.setWindowTitle("Headless sampling")

            # Add GUI elements
            vertlayout.addWidget(widget_headless)
            self.setCentralWidget(self.central_widget)
            self.central_widget.setLayout(vertlayout) # Draw elements in main widget

            # Prepare pseudo-receiver function in order to satisfy timer
            self.timer.timeout.connect(lambda: boardutils.headless_sampling())

        else:
            
            self.setWindowTitle('NeuriGUI (raw data available at {}:{})'.format(pm.udp_ip, pm.udp_port))

            # Add GUI elements
            vertlayout.addLayout(controlpanel)
            controlpanel.addWidget(widget_vrange)
            controlpanel.addWidget(widget_notch)
            controlpanel.addWidget(widget_bandpass)
            controlpanel.addWidget(widget_envelope)
            controlpanel.addWidget(widget_sbtn)
            controlpanel.addWidget(widget_darkmode)
            controlpanel.addWidget(widget_fps)
            vertlayout.addWidget(widget_signal)
            
            # Prepare data visualization
            self.timer.timeout.connect(lambda: guiwidgets.update_signal_plot(
                pm.s_down, int(pm.sample_rate * pm.buffer_add),
                pm.sample_rate,
                range(0, int(pm.sample_rate * pm.buffer_length), pm.s_down), pm.max_chans,
                [i for i in range(pm.max_chans) if pm.selected_chans[i]], shared_buffer, shared_timestamp))

        # Start sampling processes
        # -----------------------------------------------------------------
        self.sampling.start()
        self.timer.start()
        

    def on_closing(self):
        self.gui_running.value = 0
        sleep(2) # A second longer than the sampling thread
        self.timer.stop()
        self.sampling.terminate()


class Run():

    def __init__(self):

        app                     = QtWidgets.QApplication(sys.argv)
        maingui                 = MainWindow()  # Contains all necessary bits
        maingui.show()
        app.exec_()
        maingui.on_closing()
        if ( __package__ == "" or __package__ == None ):
            sys.exit()
        else:
            # sys.exit(app.exec_())
            app.quit()
    

if __name__ == '__main__': # If run from IDEs instead of as module
    Run()


from json import loads, dumps, decoder
from time import sleep, perf_counter
from numpy import expand_dims, concatenate, fromiter, append, array, zeros, reshape
try:
    from io_manager import IOManager
except:
    from .io_manager import IOManager
    

class SamplingUtilsNeuriLolinHelment():

    def __init__(self):
        pass # Necessary init function because we call attributes of this class in process_samples()
    

    def bin_to_voltage(self, s_bin, pga):
        # =================================================================
        # Convert binary into volts
        # Input
        #   s_bin       Float (binary) (crucial to be float, otherwise the 
        #               output "voltage" will be integer event if 
        #               method returns voltage as float)
        #   pga         Programmable gain (int)
        # Output
        #   - voltage   Float (microvolts)
        # =================================================================
        s_bin       = int(s_bin) # requieres int
        sign_bit    = 0
        if s_bin == 0:
            return 0
        if s_bin > 0 and s_bin <= 8388607:
            sign_bit = s_bin
        elif s_bin > 8388607 and s_bin <= 2*8388607:
            sign_bit = -2*8388607 + s_bin - 1
        else:
            print('sign_bit not assigned, returning 0')
        
        voltage = (4.5*sign_bit)/(pga*8388607.0)
        voltage = voltage * 1000000 # Convert to microvolts

        #voltage = random.randrange(-200, 200)
        return voltage
    

    def message_to_samples(self, str_message, s_chans):
        # -----------------------------------------------------------------
        # Extract samples from board
        # Input
        #   str_message (str) Raw message string coming from board
        #   s_chans     (int) How many channels shall the samples be 
        #               assigned to
        # Default in case of faulty message
        eeg_array       = expand_dims(
            array([0] * s_chans, dtype=float), # Crucial to set float
            axis=1)
        eeg_valid       = False
        #   eeg_array   Numpy array (floats), being channels by samples
        #   eeg_valid   Boolean whether workable data or not
        # -----------------------------------------------------------------
        str_message     = str_message[
            str_message.find('{'):str_message.find('}')+1]
        
        # Build Python dictionnary
        try:
            chanDict        = loads(str_message)
        except decoder.JSONDecodeError:
            # Corrupt message: not all channels present
            print("Skipped message")
            return eeg_array, eeg_valid

        if len(chanDict) != s_chans:
            return eeg_array, eeg_valid
        
        eeg_array       = expand_dims(
            fromiter(chanDict.values(), dtype=float), # Crucial to set float
            axis=1)
        eeg_valid       = True
        
        return eeg_array, eeg_valid
    

    def empty_message_queue(self, receiver):
        # TODO: Implement a cleaner way of flushing the port queue.
        # Officially supported flush functions in pyserial do not
        # seem to have effect.
        receiver.flush()
        receiver.reset_input_buffer()
        receiver.reset_output_buffer()

        # Implement dirty method:
        for _ in range(1,000):
            receiver.read(receiver.inWaiting())
        
    
    def setup_neuri_board(self, receiver, start_code):
        self.empty_message_queue(receiver)
        receiver.write(bytes(str(start_code), 'utf-8'))
        board_booting = True
        print('Board is booting up ...')
        while board_booting:
            raw_message = str(receiver.readline())
            print(raw_message)
            if 'Listening ...' in raw_message:
                receiver.write(bytes(str(start_code), 'utf-8')) # Try again
                sleep(1)
            elif '{' in raw_message and '}' in raw_message:
                print('Fully started')
                board_booting = False


    def process_samples(self, transmitter, parameter, shared_buffer,
                     shared_timestamp, gui_running):

        io_manager = IOManager(parameter)
        receiver = io_manager.set_up_serial_entry_point(
            parameter["baud_rate"], parameter["time_out"], parameter["port"])

        with receiver as r:

            # Open communication ----------------------------------------------
            sleep(1)

            self.setup_neuri_board(r, parameter["start_code"])
            if parameter["start_code"] == 2:
                s_per_buffer    = 1
                print('Ordered board to send data via USB. Switching mode ...')
            elif parameter["start_code"] == 3:
                s_per_buffer    = 10
                print('Ordered board to send data via Bluetooth. Switching mode ...')

            # Prealloate values of loop ---------------------------------------
            start_time          = int(perf_counter() * 1000)
            time_stamp_now      = int(perf_counter() * 1000) # Do NOT copy from start_time (will generate pointer)
            time_reset          = int(perf_counter() * 1000) # Do NOT copy from start_time (will generate pointer)
            sample_count        = int(0)
            
            buffer              = zeros((parameter["max_chans"],
                (parameter["buffer_length"] + parameter["buffer_add"]) * parameter["sample_rate"]))
            time_stamps         = zeros(parameter["buffer_length"] * parameter["sample_rate"], dtype=int)
            pga                 = parameter["PGA"]
            s_chans             = parameter["max_chans"]
            sampling_rate       = parameter["sample_rate"]
            saving_interval     = parameter["saving_interval"] * parameter["sample_rate"]
            update_buffer       = zeros((buffer.shape[0], buffer.shape[1]+1))
            update_times        = zeros((buffer.shape[1]+1), dtype=int)

            # Preallocate json relay message and relay connection
            relay_array         = io_manager.build_standard_relay_message(parameter["max_chans"])
            udp_ip              = parameter["udp_ip"]
            udp_port            = parameter["udp_port"]

            self.empty_message_queue(r)
                
            while gui_running.value == 1:
                
                raw_message             = str(r.readline())

                buffer_in, valid_eeg    = self.message_to_samples(raw_message, s_chans)

                if not valid_eeg:
                    # TODO: Implement an interpolation system
                    continue

                # Current timestamp -------------------------------------------
                time_stamp_now          = int(perf_counter() * 1000) - start_time
                # This will generate unchanged time_stamps for all samples of 
                # the incoming bufer (= 10 in case of bluetooth), but that is
                # not a problem

                # Each channel carries self.s_per_buffer amounts of samples 
                # (theoretically). Since parts of messages can be lost, we 
                # soft-set the amount of samples after checking for data 
                # integrity during data_checkpoint()
                for iS in range(s_per_buffer):

                    sample_count        = sample_count + 1
                    
                    sample              = buffer_in[:, iS]

                    # Convert binary to voltage values
                    for iBin in range(s_chans):
                        sample[iBin]    = self.bin_to_voltage(sample[iBin], pga)
                        relay_array["".join(["c", str(iBin+1)])] = str(sample[iBin])

                    update_buffer       = concatenate((buffer, expand_dims(sample, 1)), axis=1)
                    update_times        = append(time_stamps, time_stamp_now)

                    # Build new buffer and timestamp arrays
                    buffer              = update_buffer[:, 1:]
                    time_stamps         = update_times[1:]

                    # Make data available for downstream programs
                    transmitter.sendto(bytes(dumps(relay_array), "utf-8"), (udp_ip, udp_port))

                    # Update shared memory allocations for frontend
                    shared_buffer[:] = reshape(buffer, buffer.size) # Has left edge for filtering
                    shared_timestamp.value = time_stamp_now

                    # Write out samples to file -----------------------------------
                    if sample_count == saving_interval:

                        io_manager.calc_sample_rate(time_stamp_now, time_reset, 
                                            sampling_rate, time_stamps)

                        io_manager.master_write_data(buffer, time_stamps, 
                            saving_interval)
                        sample_count        = 0
                        time_reset          = time_stamp_now

            r.write(bytes(str(0), 'utf-8')) # Set board into standby
            sleep(1)
            return